# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import curses

import pytest

from tuir.page import Page, PageController, logged_in

try:
    from unittest import mock
except ImportError:
    import mock


def test_page_logged_in(terminal):

    page = mock.MagicMock()
    page.term = terminal

    @logged_in
    def func(_):
        raise RuntimeError()

    # Logged in runs the function
    page.reddit.is_oauth_session.return_value = True
    with pytest.raises(RuntimeError):
        func(page)
    message = 'Not logged in'.encode('utf-8')
    with pytest.raises(AssertionError):
        terminal.stdscr.subwin.addstr.assert_called_with(1, 1, message)

    # Logged out skips the function and displays a message
    page.reddit.is_oauth_session.return_value = False
    func(page)
    message = 'Not logged in'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_called_with(1, 1, message)


def test_page_unauthenticated(reddit, terminal, config, oauth):

    page = Page(reddit, terminal, config, oauth)
    page.controller = PageController(page, keymap=config.keymap)
    with mock.patch.object(page, 'refresh_content'), \
            mock.patch.object(page, 'content'),      \
            mock.patch.object(page, 'nav'),          \
            mock.patch.object(page, 'draw'):

        # Loop
        def func(_):
            page.active = False
        with mock.patch.object(page, 'controller'):
            page.controller.trigger = mock.MagicMock(side_effect=func)
            page.loop()
        assert page.draw.called

        # Quit, confirm
        terminal.stdscr.getch.return_value = ord('y')
        with mock.patch('sys.exit') as sys_exit:
            page.controller.trigger('q')
        assert sys_exit.called

        # Quit, deny
        terminal.stdscr.getch.return_value = terminal.ESCAPE
        with mock.patch('sys.exit') as sys_exit:
            page.controller.trigger('q')
        assert not sys_exit.called

        # Force quit
        terminal.stdscr.getch.return_value = terminal.ESCAPE
        with mock.patch('sys.exit') as sys_exit:
            page.controller.trigger('Q')
        assert sys_exit.called

        # Show help
        with mock.patch('subprocess.Popen') as Popen:
            page.controller.trigger('?')
        assert Popen.called

        logged_in_methods = [
            'a',  # Upvote
            'z',  # Downvote
            'd',  # Delete
            'e',  # Edit
            'i',  # Get inbox
            'C',  # Private message
        ]
        for ch in logged_in_methods:
            page.controller.trigger(ch)
            message = 'Not logged in'.encode('utf-8')
            terminal.stdscr.subwin.addstr.assert_called_with(1, 1, message)
            terminal.stdscr.subwin.addstr.reset_mock()


def test_page_authenticated(reddit, terminal, config, oauth, refresh_token):
    class TestPage(Page):
        def refresh_content(self, order=None, name=None):
            pass

    page = TestPage(reddit, terminal, config, oauth)
    page.controller = PageController(page, keymap=config.keymap)
    config.refresh_token = refresh_token

    # Login
    page.controller.trigger('u')
    assert reddit.is_oauth_session()

    # Logout
    terminal.stdscr.getch.return_value = ord('y')
    page.controller.trigger('u')
    assert not reddit.is_oauth_session()


def test_page_cycle_theme(reddit, terminal, config, oauth):

    page = Page(reddit, terminal, config, oauth)
    page.controller = PageController(page, keymap=config.keymap)

    page.term.set_theme()
    assert page.term.theme.name == 'default'

    with mock.patch.object(terminal, 'show_notification'), \
            mock.patch.object(page, 'draw'):

        # Next theme
        page.controller.trigger(curses.KEY_F3)
        assert page.term.theme.name == 'monochrome'
        terminal.show_notification.assert_called_with(
            'monochrome (built-in)', timeout=1)

        # Previous theme
        page.controller.trigger(curses.KEY_F2)
        assert page.term.theme.name == 'default'
        terminal.show_notification.assert_called_with(
            'default (built-in)', timeout=1)

        # Previous - will loop to one of the 256 color themes
        page.controller.trigger(curses.KEY_F2)
        assert page.term.theme.source in ('preset', 'installed')

        # Reset
        page.term.set_theme()

        # Will skip over any installed themes that aren't supported
        curses.has_colors.return_value = False
        page.controller.trigger(curses.KEY_F2)
        assert page.term.theme.required_colors == 0



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys

import pytest

from tuir.clipboard import copy
from tuir.exceptions import ProgramError


try:
    from unittest import mock
except ImportError:
    import mock


def test_copy_nix():

    with mock.patch('subprocess.Popen') as Popen, \
            mock.patch('subprocess.call', return_value=0) as call:

        # Mock out the subprocess calls
        p = mock.Mock()
        p.communicate = mock.Mock()
        Popen.return_value = p

        copy('test', 'xsel -b -i')
        assert Popen.call_args[0][0] == ['xsel', '-b', '-i']
        p.communicate.assert_called_with(input='test'.encode('utf-8'))

        copy('test ❤')
        assert Popen.call_args[0][0] == ['xclip', '-selection', 'clipboard']
        p.communicate.assert_called_with(input='test ❤'.encode('utf-8'))

def test_copy_darwin():
    with mock.patch('subprocess.Popen') as Popen, \
            mock.patch('subprocess.call', return_value=0) as call:

        sys.platform = 'darwin'

        p = mock.Mock()
        p.communicate = mock.Mock()
        Popen.return_value = p

        copy('test')
        assert Popen.call_args[0][0] == ['pbcopy', 'w']
        p.communicate.assert_called_with(input='test'.encode('utf-8'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
from collections import OrderedDict

import pytest

from tuir.mime_parsers import parsers, ImgurApiMIMEParser


RegexpType = type(re.compile(''))


URLS = OrderedDict([
    ('simple_png', (
        'http://www.example.com/i/image.png',  # 1. URL
        'http://www.example.com/i/image.png',  # 2. Direct media link
        'image/png')),                         # 3. MIME type of media
    ('simple_mpeg', (
        'http://www.example.com/v/video.mpeg',
        'http://www.example.com/v/video.mpeg',
        'video/mpeg')),
    ('simple_unknown', (
        'http://www.example.com/i/image',
        'http://www.example.com/i/image',
        None)),
    ('gfycat', (
        'https://gfycat.com/DeliciousUnfortunateAdouri',
        'https://giant.gfycat.com/DeliciousUnfortunateAdouri.mp4',
        'video/mp4')),
    ('youtube', (
        'https://www.youtube.com/watch?v=FjNdYp2gXRY',
        'https://www.youtube.com/watch?v=FjNdYp2gXRY',
        'video/x-youtube')),
    ('gifv', (
        'http://i.imgur.com/i/image.gifv',
        'http://i.imgur.com/i/image.mp4',
        'video/mp4')),
    ('reddit_uploads', (
        'https://i.reddituploads.com/a065472e47a4405da159189ee48bff46?fit=max&h=1536&w=1536&s=5639918a0c696b9bb3ec694dc3cf59ac',
        'https://i.reddituploads.com/a065472e47a4405da159189ee48bff46?fit=max&h=1536&w=1536&s=5639918a0c696b9bb3ec694dc3cf59ac',
        'image/jpeg')),
    ('imgur_1', (
        'http://imgur.com/yW0kbMi',
        'https://i.imgur.com/yW0kbMi.jpg',
        'image/jpeg')),
    ('imgur_2', (
        'http://imgur.com/gallery/yjP1v4B',
        'https://i.imgur.com/yjP1v4B.mp4',
        'video/mp4')),
    ('imgur_album', (
        'https://imgur.com/a/pRYEi',
        'https://i.imgur.com/tiUSazy.png https://i.imgur.com/phIsEAi.png https://i.imgur.com/TvT8V5m.png https://i.imgur.com/qETPFAi.png',
        'image/x-imgur-album')),
    ('instagram_image', (
        'https://www.instagram.com/p/BIxQ0vrBN2Y/?taken-by=kimchi_chic',
        re.compile('https://instagram(.*)\.jpg'),
        'image/jpeg')),
    ('instagram_video', (
        'https://www.instagram.com/p/BUm3cvEhFMt/',
        re.compile('https://instagram(.*)\.mp4'),
        'video/mp4')),
    ('streamable_video', (
        'https://streamable.com/vkc0y',
        re.compile('https://(.*)\.streamable\.com/video/mp4/(.*)\.mp4(.*)'),
        'video/mp4')),
    ('liveleak_video_1', (
        'https://www.liveleak.com/view?i=08b_1499296574',
        re.compile('https://cdn\.liveleak\.com/(.*)\.mp4(.*)'),
        'video/mp4')),
    ('liveleak_video_2', (
        'https://www.liveleak.com/view?i=cad_1487311993',
        re.compile('www\.youtube\.com/embed/D4GrlOMlOBY'),
        'video/x-youtube')),
    ('reddit_gif', (
        'https://v.redd.it/1j6sbv2npvd11',
        'https://v.redd.it/1j6sbv2npvd11/DASH_2_4_M',
        'video/mp4')),
    ('reddit_video', (
        'https://v.redd.it/zv89llsvexdz',
        'https://v.redd.it/zv89llsvexdz/DASHPlaylist.mpd',
        'video/x-youtube')),
    ('clippituser', (
        'https://www.clippituser.tv/c/edqqld',
        'https://clips.clippit.tv/edqqld/360.mp4',
        'video/mp4')),
    ('gifs_com', (
        'https://gifs.com/gif/cat-jumps-into-lake-after-guy-sneezes-0VqvRL',
        'https://j.gifs.com/0VqvRL.gif',
        'image/gif')),
    ('giphy', (
        'https://giphy.com/gifs/cameron-dallas-OpesLQSjwdGj6',
        'https://media.giphy.com/media/OpesLQSjwdGj6/giphy.mp4',
        'video/mp4')),
    ('imgflip', (
        'https://imgflip.com/i/1dtdbv',
        'https://i.imgflip.com/1dtdbv.jpg',
        'image/jpeg')),
    ('livememe', (
        'http://www.livememe.com/v67cxy9',
        'http://e.lvme.me/v67cxy9.jpg',
        'image/jpeg')),
    ('makeameme', (
        'http://makeameme.org/meme/joined-reddit-for',
        'https://media.makeameme.org/created/joined-reddit-for.jpg',
        'image/jpeg')),
    ('flickr', (
        'https://www.flickr.com/photos/obamawhitehouse/8191317327/',
        'https://c1.staticflickr.com/9/8065/8191317327_5180e95d98_b.jpg',
        'image/jpeg')),
    ('worldstar_1', (
        'http://www.worldstarhiphop.com/videos/video.php?v=wshh09DUN5MeDK2El4sO',
        'https://hw-videos.worldstarhiphop.com/u/vid/2017/10/4OnndgEmNSME.mp4',
        'video/mp4')),
    ('worldstar_2', (
        'http://www.worldstarhiphop.com/videos/video.php?v=wshhJ6bVdAv0iMrNGFZG',
        'https://www.youtube.com/embed/Bze53qwHS8o?rel=0&autoplay=1',
        'video/x-youtube')),
    ('vimeo', (
        'https://vimeo.com/247872788',
        'https://vimeo.com/247872788',
        'video/x-youtube')),
    ('streamja', (
        'https://streamja.com/qLaQ',
        'https://cdnja.r.worldssl.net/mp4/ql/qLaQ.mp4',
        'video/mp4')),
])


args, ids = URLS.values(), list(URLS)
@pytest.mark.parametrize('url,modified_url,mime_type', args, ids=ids)
def test_parser(url, modified_url, mime_type, reddit, config):
    # Include the reddit fixture so the cassettes get generated

    ImgurApiMIMEParser.CLIENT_ID = config['imgur_client_id']

    for parser in parsers:
        if parser.pattern.match(url):
            parsed_url, parsed_type = parser.get_mimetype(url)
            if isinstance(modified_url, RegexpType):
                assert modified_url.match(parsed_url)
            else:
                assert modified_url == parsed_url
            assert parsed_type == mime_type
            break
    else:
        # The base parser should catch all urls before this point
        assert False


def test_imgur_fallback(reddit):
    """
    If something happens to the imgur API key, the code should fallback
    to manually scraping the page.
    """

    ImgurApiMIMEParser.CLIENT_ID = ''
    for key in ['imgur_1', 'imgur_2']:
        url, modified_url, mime_type = URLS[key]

        assert ImgurApiMIMEParser.pattern.match(url)
        parsed_url, parsed_type = ImgurApiMIMEParser.get_mimetype(url)
        # Not sure why, but http://imgur.com/gallery/yjP1v4B (a .gif)
        # appears to incorrectly return as a JPG type from the scraper
        assert parsed_type is not None

    # The fallback scraper for albums no longer exists, but it should
    # return the original URL instead of raising an error
    url, modified_url, mime_type = URLS['imgur_album']
    assert ImgurApiMIMEParser.pattern.match(url)
    parsed_url, parsed_type = ImgurApiMIMEParser.get_mimetype(url)
    assert parsed_type is None
    assert parsed_url == url

    ImgurApiMIMEParser.CLIENT_ID = 'invalid_api_key'
    for key in ['imgur_1', 'imgur_2']:
        url, modified_url, mime_type = URLS[key]

        assert ImgurApiMIMEParser.pattern.match(url)
        parsed_url, parsed_type = ImgurApiMIMEParser.get_mimetype(url)
        # Not sure why, but http://imgur.com/gallery/yjP1v4B (a .gif)
        # appears to incorrectly return as a JPG type from the scraper
        assert parsed_type is not None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import curses
from collections import OrderedDict

import pytest

from tuir.submission_page import SubmissionPage
from tuir.docs import FOOTER_SUBMISSION

try:
    from unittest import mock
except ImportError:
    import mock


PROMPTS = OrderedDict([
    ('prompt_1', 'comments/571dw3'),
    ('prompt_2', '///comments/571dw3'),
    ('prompt_3', '/comments/571dw3'),
    ('prompt_4', '/r/pics/comments/571dw3/'),
    ('prompt_5', 'https://www.reddit.com/r/pics/comments/571dw3/at_disneyland'),
])


def test_submission_page_construct(reddit, terminal, config, oauth):
    window = terminal.stdscr.subwin
    url = ('https://www.reddit.com/r/Python/comments/2xmo63/'
           'a_python_terminal_viewer_for_browsing_reddit')

    with terminal.loader():
        page = SubmissionPage(reddit, terminal, config, oauth, url=url)
    assert terminal.loader.exception is None

    # Toggle the second comment so we can check the draw more comments method
    page.content.toggle(1)

    # Set some special flags to make sure that we can draw them
    submission_data = page.content.get(-1)
    submission_data['gold'] = 1
    submission_data['stickied'] = True
    submission_data['saved'] = True
    submission_data['flair'] = 'flair'

    # Set some special flags to make sure that we can draw them
    comment_data = page.content.get(0)
    comment_data['gold'] = 3
    comment_data['stickied'] = True
    comment_data['saved'] = True
    comment_data['flair'] = 'flair'

    page.draw()

    #  Title
    title = url[:terminal.stdscr.ncols-1].encode('utf-8')
    window.addstr.assert_any_call(0, 0, title)

    # Banner
    menu = '[1]hot         [2]top         [3]rising         [4]new         [5]controversial'
    window.addstr.assert_any_call(0, 0, menu.encode('utf-8'))

    # Footer - The text is longer than the default terminal width
    text = FOOTER_SUBMISSION.strip()[:79]
    window.addstr.assert_any_call(0, 0, text.encode('utf-8'))

    # Submission
    submission_data = page.content.get(-1)
    text = submission_data['title'].encode('utf-8')
    window.subwin.addstr.assert_any_call(1, 1, text, 2097152)
    assert window.subwin.border.called

    # Comment
    comment_data = page.content.get(0)
    text = comment_data['split_body'][0].encode('utf-8')
    window.subwin.addstr.assert_any_call(1, 1, text, curses.A_NORMAL)

    # More Comments
    comment_data = page.content.get(1)
    text = comment_data['body'].encode('utf-8')
    window.subwin.addstr.assert_any_call(0, 1, text, curses.A_NORMAL)

    # Cursor should not be drawn when the page is first opened
    assert not any(args[0][3] == curses.A_REVERSE
                   for args in window.subwin.addch.call_args_list)

    # Reload with a smaller terminal window
    terminal.stdscr.ncols = 20
    terminal.stdscr.nlines = 10
    with terminal.loader():
        page = SubmissionPage(reddit, terminal, config, oauth, url=url)
    assert terminal.loader.exception is None
    page.draw()


def test_submission_refresh(submission_page):

    # Should be able to refresh content
    submission_page.refresh_content()


def test_submission_exit(submission_page):

    # Exiting should set active to false
    submission_page.active = True
    submission_page.controller.trigger('h')
    assert not submission_page.active


def test_submission_unauthenticated(submission_page, terminal):

    # Unauthenticated commands
    methods = [
        'a',  # Upvote
        'z',  # Downvote
        'c',  # Comment
        'e',  # Edit
        'd',  # Delete
        'w',  # Save
    ]
    for ch in methods:
        submission_page.controller.trigger(ch)
        text = 'Not logged in'.encode('utf-8')
        terminal.stdscr.subwin.addstr.assert_called_with(1, 1, text)


def test_submission_open(submission_page, terminal):

    # Open the selected link with the web browser
    with mock.patch.object(terminal, 'open_browser'):
        submission_page.controller.trigger(terminal.RETURN)
        assert terminal.open_browser.called


def test_submission_prompt(submission_page, terminal):

    # Prompt for a different subreddit
    with mock.patch.object(terminal, 'prompt_input'):
        # Valid input
        submission_page.active = True
        submission_page.selected_page = None
        terminal.prompt_input.return_value = 'front/top'
        submission_page.controller.trigger('/')

        submission_page.handle_selected_page()
        assert not submission_page.active
        assert submission_page.selected_page

        # Invalid input
        submission_page.active = True
        submission_page.selected_page = None
        terminal.prompt_input.return_value = 'front/pot'
        submission_page.controller.trigger('/')

        submission_page.handle_selected_page()
        assert submission_page.active
        assert not submission_page.selected_page


@pytest.mark.parametrize('prompt', PROMPTS.values(), ids=list(PROMPTS))
def test_submission_prompt_submission(submission_page, terminal, prompt):

    # Navigate to a different submission from inside a submission
    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = prompt
        submission_page.content.order = 'top'
        submission_page.controller.trigger('/')
        assert not terminal.loader.exception

        submission_page.handle_selected_page()
        assert not submission_page.active
        assert submission_page.selected_page

        assert submission_page.selected_page.content.order is None
        data = submission_page.selected_page.content.get(-1)
        assert data['object'].id == '571dw3'


def test_submission_order(submission_page):

    submission_page.controller.trigger('1')
    assert submission_page.content.order == 'hot'
    submission_page.controller.trigger('2')
    assert submission_page.content.order == 'top'
    submission_page.controller.trigger('3')
    assert submission_page.content.order == 'rising'
    submission_page.controller.trigger('4')
    assert submission_page.content.order == 'new'
    submission_page.controller.trigger('5')
    assert submission_page.content.order == 'controversial'

    # Shouldn't be able to sort the submission page by gilded
    submission_page.controller.trigger('6')
    assert submission_page.content.order == 'controversial'


def test_submission_move_top_bottom(submission_page):

    submission_page.controller.trigger('G')
    assert submission_page.nav.absolute_index == 44

    submission_page.controller.trigger('g')
    submission_page.controller.trigger('g')
    assert submission_page.nav.absolute_index == -1


def test_submission_move_sibling_parent(submission_page):

    # Jump to sibling
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')
        submission_page.controller.trigger('J')
    assert submission_page.nav.absolute_index == 7

    # Jump to parent
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('k')
        submission_page.controller.trigger('k')
        submission_page.controller.trigger('K')
    assert submission_page.nav.absolute_index == 0


def test_submission_pager(submission_page, terminal):

    # View a submission with the pager
    with mock.patch.object(terminal, 'open_pager'):
        submission_page.controller.trigger('l')
        assert terminal.open_pager.called

    # Move down to the first comment
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # View a comment with the pager
    with mock.patch.object(terminal, 'open_pager'):
        submission_page.controller.trigger('l')
        assert terminal.open_pager.called


def test_submission_comment_not_enough_space(submission_page, terminal):

    # The first comment is 10 lines, shrink the screen so that it won't fit.
    # Setting the terminal to 10 lines means that there will only be 8 lines
    # available (after subtracting the header and footer) to draw the comment.
    terminal.stdscr.nlines = 10

    # Select the first comment
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.move_cursor_down()

    submission_page.draw()

    text = '(Not enough space to display)'.encode('ascii')
    window = terminal.stdscr.subwin
    window.subwin.addstr.assert_any_call(6, 1, text, curses.A_NORMAL)


def test_submission_vote(submission_page, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Test voting on the submission
    with mock.patch('tuir.packages.praw.objects.Submission.upvote') as upvote,            \
            mock.patch('tuir.packages.praw.objects.Submission.downvote') as downvote,     \
            mock.patch('tuir.packages.praw.objects.Submission.clear_vote') as clear_vote:

        data = submission_page.get_selected_item()
        data['object'].archived = False

        # Upvote
        submission_page.controller.trigger('a')
        assert upvote.called
        assert data['likes'] is True

        # Clear vote
        submission_page.controller.trigger('a')
        assert clear_vote.called
        assert data['likes'] is None

        # Upvote
        submission_page.controller.trigger('a')
        assert upvote.called
        assert data['likes'] is True

        # Downvote
        submission_page.controller.trigger('z')
        assert downvote.called
        assert data['likes'] is False

        # Clear vote
        submission_page.controller.trigger('z')
        assert clear_vote.called
        assert data['likes'] is None

        # Upvote - exception
        upvote.side_effect = KeyboardInterrupt
        submission_page.controller.trigger('a')
        assert data['likes'] is None

        # Downvote - exception
        downvote.side_effect = KeyboardInterrupt
        submission_page.controller.trigger('a')
        assert data['likes'] is None


def test_submission_vote_archived(submission_page, refresh_token, terminal):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Load an archived submission
    archived_url = 'https://www.reddit.com/r/IAmA/comments/z1c9z/'
    submission_page.refresh_content(name=archived_url)

    with mock.patch.object(terminal, 'show_notification') as show_notification:
        data = submission_page.get_selected_item()

        # Upvote the submission
        show_notification.reset_mock()
        submission_page.controller.trigger('a')
        show_notification.assert_called_with('Voting disabled for archived post', style='Error')
        assert data['likes'] is None

        # Downvote the submission
        show_notification.reset_mock()
        submission_page.controller.trigger('z')
        show_notification.assert_called_with('Voting disabled for archived post', style='Error')
        assert data['likes'] is None


def test_submission_save(submission_page, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Test save on the submission
    with mock.patch('tuir.packages.praw.objects.Submission.save') as save,        \
            mock.patch('tuir.packages.praw.objects.Submission.unsave') as unsave:

        data = submission_page.content.get(submission_page.nav.absolute_index)

        # Save
        submission_page.controller.trigger('w')
        assert save.called
        assert data['saved'] is True

        # Unsave
        submission_page.controller.trigger('w')
        assert unsave.called
        assert data['saved'] is False

        # Save - exception
        save.side_effect = KeyboardInterrupt
        submission_page.controller.trigger('w')
        assert data['saved'] is False


def test_submission_comment_save(submission_page, terminal, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Move down to the first comment
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # Test save on the comment submission
    with mock.patch('tuir.packages.praw.objects.Comment.save') as save,        \
            mock.patch('tuir.packages.praw.objects.Comment.unsave') as unsave:

        data = submission_page.content.get(submission_page.nav.absolute_index)

        # Save
        submission_page.controller.trigger('w')
        assert save.called
        assert data['saved'] is True

        # Unsave
        submission_page.controller.trigger('w')
        assert unsave.called
        assert data['saved'] is False

        # Save - exception
        save.side_effect = KeyboardInterrupt
        submission_page.controller.trigger('w')
        assert data['saved'] is False


def test_submission_comment(submission_page, terminal, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Leave a comment
    with mock.patch('tuir.packages.praw.objects.Submission.add_comment') as add_comment, \
            mock.patch.object(terminal, 'open_editor') as open_editor,                  \
            mock.patch('time.sleep'):
        open_editor.return_value.__enter__.return_value = 'comment text'
        submission_page.controller.trigger('c')
        assert open_editor.called
        add_comment.assert_called_with('comment text')


def test_submission_delete(submission_page, terminal, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Can't delete the submission
    curses.flash.reset_mock()
    submission_page.controller.trigger('d')
    assert curses.flash.called

    # Move down to the first comment
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # Try to delete the first comment - wrong author
    curses.flash.reset_mock()
    submission_page.controller.trigger('d')
    assert curses.flash.called

    # Spoof the author and try to delete again
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['author'] = submission_page.reddit.user.name
    with mock.patch('tuir.packages.praw.objects.Comment.delete') as delete,  \
            mock.patch.object(terminal.stdscr, 'getch') as getch,           \
            mock.patch('time.sleep'):
        getch.return_value = ord('y')
        submission_page.controller.trigger('d')
        assert delete.called


def test_submission_edit(submission_page, terminal, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Try to edit the submission - wrong author
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['author'] = 'some other person'
    curses.flash.reset_mock()
    submission_page.controller.trigger('e')
    assert curses.flash.called

    # Spoof the submission and submit empty comment
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['author'] = submission_page.reddit.user.name
    with mock.patch('tuir.packages.praw.objects.Submission.edit') as edit,  \
            mock.patch.object(terminal, 'open_editor') as open_editor,     \
            mock.patch.object(terminal, 'show_notification'),  \
            mock.patch('time.sleep'):
        open_editor.return_value.__enter__.return_value = ''

        submission_page.controller.trigger('e')
        assert open_editor.called
        terminal.show_notification.assert_called_with('Canceled')
        edit.assert_not_called()

    # Spoof the submission and try to edit again
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['author'] = submission_page.reddit.user.name
    with mock.patch('tuir.packages.praw.objects.Submission.edit') as edit,  \
            mock.patch.object(terminal, 'open_editor') as open_editor,     \
            mock.patch('time.sleep'):
        open_editor.return_value.__enter__.return_value = 'submission text'

        submission_page.controller.trigger('e')
        assert open_editor.called
        edit.assert_called_with('submission text')

    # Move down to the first comment
    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # Spoof the author and edit the comment
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['author'] = submission_page.reddit.user.name
    with mock.patch('tuir.packages.praw.objects.Comment.edit') as edit, \
            mock.patch.object(terminal, 'open_editor') as open_editor, \
            mock.patch('time.sleep'):
        open_editor.return_value.__enter__.return_value = 'comment text'

        submission_page.controller.trigger('e')
        assert open_editor.called
        edit.assert_called_with('comment text')


def test_submission_urlview(submission_page, terminal, refresh_token):

    # Log in
    submission_page.config.refresh_token = refresh_token
    submission_page.oauth.authorize()

    # Submission case
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['body'] = 'test comment body  ❤'
    with mock.patch.object(terminal, 'open_urlview') as open_urlview:
        submission_page.controller.trigger('b')
        open_urlview.assert_called_with('test comment body  ❤')

    # Subreddit case
    data = submission_page.content.get(submission_page.nav.absolute_index)
    data['text'] = ''
    data['body'] = ''
    data['url_full'] = 'http://test.url.com  ❤'
    with mock.patch.object(terminal, 'open_urlview') as open_urlview, \
            mock.patch('subprocess.Popen'):
        submission_page.controller.trigger('b')
        open_urlview.assert_called_with('http://test.url.com  ❤')


def test_submission_prompt_and_select_link(submission_page, terminal):

    # A link submission should return the URL that it's pointing to
    link = submission_page.prompt_and_select_link()
    assert link == 'https://github.com/michael-lazar/rtv'

    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # The first comment doesn't have any links in the comment body
    link = submission_page.prompt_and_select_link()
    data = submission_page.get_selected_item()
    assert link == data['permalink']

    with mock.patch.object(submission_page, 'clear_input_queue'):
        submission_page.controller.trigger('j')

    # The second comment has a link embedded in the comment body, and
    # the user is prompted to select which link to open
    with mock.patch.object(terminal, 'prompt_user_to_select_link') as prompt:
        prompt.return_value = 'https://selected_link'

        link = submission_page.prompt_and_select_link()
        data = submission_page.get_selected_item()

        assert link == prompt.return_value

        embedded_url = 'http://peterdowns.com/posts/first-time-with-pypi.html'
        assert prompt.call_args[0][0] == [
            {'text': 'Permalink', 'href': data['permalink']},
            {'text': 'Relevant tutorial', 'href': embedded_url}
        ]

    submission_page.controller.trigger(' ')

    # The comment is now hidden so there are no links to select
    link = submission_page.prompt_and_select_link()
    assert link is None


def test_submission_copy_permalink(submission_page, terminal):
    url = ('https://www.reddit.com/r/Python/comments/2xmo63/'
           'a_python_terminal_viewer_for_browsing_reddit/')
    with mock.patch('tuir.page.clipboard_copy') as copy, \
            mock.patch.object(terminal, 'show_notification'):
        submission_page.controller.trigger('y')
        copy.assert_called_with(url, mock.ANY)
        terminal.show_notification.assert_called_with(['Copied to clipboard:', url], timeout=1)


def test_submission_copy_link(submission_page, terminal):
    url = 'https://github.com/michael-lazar/rtv'
    with mock.patch('tuir.page.clipboard_copy') as copy, \
            mock.patch.object(terminal, 'prompt_user_to_select_link') as prompt, \
            mock.patch.object(terminal, 'show_notification'):
        submission_page.controller.trigger('Y')
        copy.assert_called_with(url, mock.ANY)
        terminal.show_notification.assert_called_with(['Copied to clipboard:', url], timeout=1)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import codecs
from tempfile import NamedTemporaryFile

from tuir.config import Config, copy_default_config, copy_default_mailcap, _copy_settings_file

try:
    from unittest import mock
except ImportError:
    import mock


# Don't need to test the functionality of _copy_settings_file in these two,
# just that they call _copy_settings_file
def test_copy_default_mailcap():
    """Make sure the default config file was included in the package"""

    with NamedTemporaryFile() as fp:
        with mock.patch('tuir.config.six.moves.input', return_value='y'), \
             mock.patch('tuir.config._copy_settings_file') as _copy_settings_file:

            copy_default_mailcap(fp.name)
            _copy_settings_file.assert_called_with(Config.DEFAULT_MAILCAP, fp.name, 'mailcap')


def test_copy_default_config():
    """Make sure the default config file was included in the package"""

    with NamedTemporaryFile(suffix='.cfg') as fp:
        with mock.patch('tuir.config.six.moves.input', return_value='y'), \
             mock.patch('tuir.config._copy_settings_file') as _copy_settings_file:

            copy_default_config(fp.name)
            _copy_settings_file.assert_called_with(Config.DEFAULT_CONFIG, fp.name, 'config')


def test_copy_default_config_cancel():
    """Pressing ``n`` should cancel the copy"""

    with NamedTemporaryFile(suffix='.cfg') as fp:
        with mock.patch('tuir.config.six.moves.input', return_value='n'):
            copy_default_config(fp.name)
            assert not fp.read()


def test_copy_config_interrupt():
    """Pressing ``Ctrl-C`` should cancel the copy"""

    with NamedTemporaryFile(suffix='.cfg') as fp:
        with mock.patch('tuir.config.six.moves.input') as func:
            func.side_effect = KeyboardInterrupt
            copy_default_config(fp.name)
            assert not fp.read()


def test__copy_settings_file():
    with NamedTemporaryFile() as fp:
        with mock.patch('tuir.config.six.moves.input', return_value='y'):
            # Doesn't matter what values we actually use, as long as they
            # are sane
            _copy_settings_file(Config.DEFAULT_CONFIG, fp.name, 'config')
            assert fp.read()

            # Check that the permissions were changed
            permissions = os.stat(fp.name).st_mode & 0o777
            assert permissions == 0o664


def test__copy_settings_file_makedirs():
    with NamedTemporaryFile() as fp:
        with mock.patch('tuir.config.six.moves.input', return_value='y'), \
             mock.patch('os.path.exists', return_value=False),            \
             mock.patch('os.makedirs'):

            _copy_settings_file(Config.DEFAULT_CONFIG, fp.name, 'config')
            assert os.makedirs.called


def test_config_interface():
    """Test setting and removing values"""

    config = Config(ascii=True)
    assert config['ascii'] is True
    config['ascii'] = False
    assert config['ascii'] is False
    config['ascii'] = None
    assert config['ascii'] is None
    del config['ascii']
    assert config['ascii'] is False

    config.update(subreddit='cfb', new_value=2.0)
    assert config['subreddit'] == 'cfb'
    assert config['new_value'] == 2.0

    assert config['link'] is None
    assert config['log'] is None


def test_config_get_args():
    """Ensure that command line arguments are parsed properly"""

    args = ['tuir',
            'https://reddit.com/permalink •',
            '-s', 'cfb',
            '--log', 'logfile.log',
            '--config', 'configfile.cfg',
            '--ascii',
            '--monochrome',
            '--non-persistent',
            '--clear-auth',
            '--copy-config',
            '--enable-media',
            '--theme', 'molokai',
            '--list-themes',
            '--no-flash',
            '--no-autologin']

    with mock.patch('sys.argv', ['tuir']):
        config_dict = Config.get_args()
        config = Config(**config_dict)
        assert config.config == {}

    with mock.patch('sys.argv', args):
        config_dict = Config.get_args()

        config = Config(**config_dict)
        assert config['ascii'] is True
        assert config['monochrome'] is True
        assert config['subreddit'] == 'cfb'
        assert config['log'] == 'logfile.log'
        assert config['ascii'] is True
        assert config['persistent'] is False
        assert config['clear_auth'] is True
        assert config['link'] == 'https://reddit.com/permalink •'
        assert config['config'] == 'configfile.cfg'
        assert config['copy_config'] is True
        assert config['enable_media'] is True
        assert config['theme'] == 'molokai'
        assert config['list_themes'] is True
        assert config['flash'] is False
        assert config['autologin'] is False


def test_config_link_deprecated():

    # Should still be able to specify the link using the old "-l"
    args = ['tuir', '-l', 'https://reddit.com/option']
    with mock.patch('sys.argv', args):
        config_dict = Config.get_args()
        config = Config(**config_dict)
        assert config['link'] == 'https://reddit.com/option'

    # But the positional argument should take preference
    args = ['tuir', 'https://reddit.com/arg', '-l', 'https://reddit.com/option']
    with mock.patch('sys.argv', args):
        config_dict = Config.get_args()
        config = Config(**config_dict)
        assert config['link'] == 'https://reddit.com/arg'


def test_config_from_file():
    """Ensure that config file arguments are parsed properly"""

    args = {
        'ascii': True,
        'monochrome': True,
        'persistent': False,
        'clear_auth': True,
        'log': 'logfile.log',
        'link': 'https://reddit.com/permalink •',
        'subreddit': 'cfb',
        'enable_media': True,
        'max_comment_cols': 150,
        'max_pager_cols': 120,
        'hide_username': True,
        'theme': 'molokai',
        'flash': True,
        'autologin': True,
        'copy_as_old': False,
    }

    bindings = {
        'REFRESH': 'r, <KEY_F5>',
        'UPVOTE': ''
    }

    with NamedTemporaryFile(suffix='.cfg') as fp:

        fargs, fbindings = Config.get_file(filename=fp.name)
        config = Config(**fargs)
        default_keymap = config.keymap._keymap.copy()
        config.keymap.set_bindings(fbindings)
        assert config.config == {}
        assert config.keymap._keymap == default_keymap

        # [tuir]
        rows = ['{0}={1}'.format(key, val) for key, val in args.items()]
        data = '\n'.join(['[tuir]'] + rows)
        fp.write(codecs.encode(data, 'utf-8'))

        # [bindings]
        rows = ['{0}={1}'.format(key, val) for key, val in bindings.items()]
        data = '\n'.join(['', '', '[bindings]'] + rows)
        fp.write(codecs.encode(data, 'utf-8'))

        fp.flush()
        fargs, fbindings = Config.get_file(filename=fp.name)
        config.update(**fargs)
        config.keymap.set_bindings(fbindings)
        assert config.config == args
        assert config.keymap.get('REFRESH') == ['r', '<KEY_F5>']
        assert config.keymap.get('UPVOTE') == ['']


def test_config_refresh_token():
    """Ensure that the refresh token can be loaded, saved, and removed"""

    with NamedTemporaryFile(delete=False) as fp:
        config = Config(token_file=fp.name)

        # Write a new token to the file
        config.refresh_token = 'secret_value'
        config.save_refresh_token()

        # Load a valid token from the file
        config.refresh_token = None
        config.load_refresh_token()
        assert config.refresh_token == 'secret_value'

        # Discard the token and delete the file
        config.delete_refresh_token()
        config.delete_refresh_token()
        assert config.refresh_token is None
        assert not os.path.exists(fp.name)

        # Saving should create a new file
        config.refresh_token = 'new_value'
        config.save_refresh_token()

        # Which we can read back to verify
        config.refresh_token = None
        config.load_refresh_token()
        assert config.refresh_token == 'new_value'

        # And delete again to clean up
        config.delete_refresh_token()
        assert not os.path.exists(fp.name)

        # Loading from the non-existent file should return None
        config.refresh_token = 'secret_value'
        config.load_refresh_token()
        assert config.refresh_token is None


def test_config_history():
    """Ensure that the history can be loaded and saved"""

    # Should still be able to load if the file doesn't exist
    config = Config(history_file='/fake_path/fake_file')
    config.load_history()
    assert len(config.history) == 0

    with NamedTemporaryFile(delete=False) as fp:
        config = Config(history_file=fp.name, history_size=3)

        config.history.add('link1')
        config.history.add('link2')
        config.history.add('link3')
        config.history.add('link4')
        assert len(config.history) == 4

        # Saving should only write the 3 most recent links
        config.save_history()
        config.load_history()
        assert len(config.history) == 3
        assert 'link1' not in config.history
        assert 'link4' in config.history

        config.delete_history()
        assert len(config.history) == 0
        assert not os.path.exists(fp.name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
import os
import curses
import codecs
from textwrap import dedent

import six
import pytest

from tuir.theme import Theme
from tuir.terminal import Terminal
from tuir.docs import (HELP, REPLY_FILE, COMMENT_EDIT_FILE, TOKEN,
                      SUBMISSION_FILE, SUBMISSION_EDIT_FILE, MESSAGE_FILE)
from tuir.exceptions import TemporaryFileError, BrowserError

try:
    from unittest import mock
except ImportError:
    import mock


def test_terminal_properties(terminal, config):

    assert isinstance(terminal.up_arrow, six.text_type)
    assert isinstance(terminal.down_arrow, six.text_type)
    assert isinstance(terminal.neutral_arrow, six.text_type)
    assert isinstance(terminal.gilded, six.text_type)

    terminal._display = None
    with mock.patch('tuir.terminal.sys') as sys, \
            mock.patch.dict('os.environ', {'DISPLAY': ''}):
        sys.platform = 'linux'
        assert terminal.display is False

    terminal._display = None
    with mock.patch('tuir.terminal.sys') as sys, \
            mock.patch('os.environ', {'DISPLAY': ''}), \
            mock.patch('webbrowser._tryorder', new=[]):
        sys.platform = 'darwin'
        assert terminal.display is True

    terminal._display = None
    with mock.patch.dict('os.environ', {'DISPLAY': ':0', 'BROWSER': 'w3m'}):
        assert terminal.display is False

    terminal._display = None
    with mock.patch.dict('os.environ', {'DISPLAY': ':0', 'BROWSER': ''}), \
            mock.patch('webbrowser._tryorder'):
        assert terminal.display is True

    assert terminal.get_arrow(None) is not None
    assert terminal.get_arrow(True) is not None
    assert terminal.get_arrow(False) is not None
    assert terminal.config == config
    assert terminal.loader is not None

    assert terminal.MIN_HEIGHT is not None
    assert terminal.MIN_WIDTH is not None

    assert terminal.theme is not None


def test_terminal_functions(terminal):

    terminal.flash()
    assert curses.flash.called

    terminal.getch()
    assert terminal.stdscr.getch.called

    with pytest.raises(RuntimeError):
        with terminal.no_delay():
            raise RuntimeError()
    terminal.stdscr.nodelay.assert_any_call(0)
    terminal.stdscr.nodelay.assert_any_call(1)

    curses.endwin.reset_mock()
    curses.doupdate.reset_mock()
    with terminal.suspend():
        pass
    assert curses.endwin.called
    assert curses.doupdate.called

    curses.endwin.reset_mock()
    curses.doupdate.reset_mock()
    with pytest.raises(RuntimeError):
        with terminal.suspend():
            raise RuntimeError()
    assert curses.endwin.called
    assert curses.doupdate.called

    terminal.addch(terminal.stdscr, 3, 5, 'ch', 'attr')
    terminal.stdscr.addch.assert_called_with(3, 5, 'ch', 'attr')


def test_terminal_no_flash(terminal):

    terminal.config['flash'] = False
    terminal.flash()
    assert not curses.flash.called


def test_terminal_clean_ascii(terminal):

    terminal.config['ascii'] = True

    # unicode returns ascii
    text = terminal.clean('hello ❤')
    assert isinstance(text, six.binary_type)
    assert text.decode('ascii') == 'hello ?'

    # utf-8 returns ascii
    text = terminal.clean('hello ❤'.encode('utf-8'))
    assert isinstance(text, six.binary_type)
    assert text.decode('ascii') == 'hello ?'

    # ascii returns ascii
    text = terminal.clean('hello'.encode('ascii'))
    assert isinstance(text, six.binary_type)
    assert text.decode('ascii') == 'hello'


def test_terminal_clean_unicode(terminal):

    terminal.config['ascii'] = False

    # unicode returns utf-8
    text = terminal.clean('hello ❤')
    assert isinstance(text, six.binary_type)
    assert text.decode('utf-8') == 'hello ❤'

    # utf-8 returns utf-8
    text = terminal.clean('hello ❤'.encode('utf-8'))
    assert isinstance(text, six.binary_type)
    assert text.decode('utf-8') == 'hello ❤'

    # ascii returns utf-8
    text = terminal.clean('hello'.encode('ascii'))
    assert isinstance(text, six.binary_type)
    assert text.decode('utf-8') == 'hello'


def test_terminal_clean_ncols(terminal):

    text = terminal.clean('hello', n_cols=5)
    assert text.decode('utf-8') == 'hello'

    text = terminal.clean('hello', n_cols=4)
    assert text.decode('utf-8') == 'hell'

    text = terminal.clean('ｈｅｌｌｏ', n_cols=10)
    assert text.decode('utf-8') == 'ｈｅｌｌｏ'

    text = terminal.clean('ｈｅｌｌｏ', n_cols=9)
    assert text.decode('utf-8') == 'ｈｅｌｌ'


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_clean_unescape_html(terminal, use_ascii):

    # HTML characters get decoded
    terminal.config['ascii'] = use_ascii
    text = terminal.clean('&lt;')
    assert isinstance(text, six.binary_type)
    assert text.decode('ascii' if use_ascii else 'utf-8') == '<'


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_add_line(terminal, stdscr, use_ascii):

    terminal.config['ascii'] = use_ascii

    terminal.add_line(stdscr, 'hello')
    print(stdscr.method_calls)
    stdscr.addstr.assert_called_with(0, 0, 'hello'.encode('ascii'))
    stdscr.reset_mock()

    # Text will be drawn, but cut off to fit on the screen
    terminal.add_line(stdscr, 'hello', row=3, col=75)
    print(stdscr.method_calls)
    stdscr.addstr.assert_called_with(3, 75, 'hell'.encode('ascii'))
    stdscr.reset_mock()

    # Outside of screen bounds, don't even try to draw the text
    terminal.add_line(stdscr, 'hello', col=79)
    assert not stdscr.addstr.called
    stdscr.reset_mock()


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_show_notification(terminal, stdscr, use_ascii):

    terminal.config['ascii'] = use_ascii

    # Multi-line messages should be automatically split
    text = 'line 1\nline 2\nline3'
    terminal.show_notification(text)
    assert stdscr.subwin.nlines == 5
    assert stdscr.subwin.addstr.call_count == 3
    stdscr.reset_mock()

    # The text should be trimmed to fit 40x80
    text = HELP.strip().splitlines()
    terminal.show_notification(text)
    assert stdscr.subwin.nlines == 40
    assert stdscr.subwin.ncols <= 80
    assert stdscr.subwin.addstr.call_count == 38
    stdscr.reset_mock()

    # The text should be trimmed to fit in 20x20
    stdscr.nlines, stdscr.ncols = 15, 20
    text = HELP.strip().splitlines()
    terminal.show_notification(text)
    assert stdscr.subwin.nlines == 15
    assert stdscr.subwin.ncols == 20
    assert stdscr.subwin.addstr.call_count == 13


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_text_input(terminal, stdscr, use_ascii):

    terminal.config['ascii'] = use_ascii
    stdscr.nlines = 1

    # Text will be wrong because stdscr.inch() is not implemented
    # But we can at least tell if text was captured or not
    stdscr.getch.side_effect = [ord('h'), ord('i'), ord('!'), terminal.RETURN]
    assert isinstance(terminal.text_input(stdscr), six.text_type)

    stdscr.getch.side_effect = [ord('b'), ord('y'), ord('e'), terminal.ESCAPE]
    assert terminal.text_input(stdscr) is None

    stdscr.getch.side_effect = [ord('h'), curses.KEY_RESIZE, terminal.RETURN]
    assert terminal.text_input(stdscr, allow_resize=True) is not None

    stdscr.getch.side_effect = [ord('h'), curses.KEY_RESIZE, terminal.RETURN]
    assert terminal.text_input(stdscr, allow_resize=False) is None


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_prompt_input(terminal, stdscr, use_ascii):

    terminal.config['ascii'] = use_ascii
    window = stdscr.derwin()

    window.getch.side_effect = [ord('h'), ord('i'), terminal.RETURN]
    assert isinstance(terminal.prompt_input('hi'), six.text_type)

    stdscr.subwin.addstr.assert_called_with(0, 0, 'hi'.encode('ascii'))
    assert window.nlines == 1
    assert window.ncols == 78

    window.getch.side_effect = [ord('b'), ord('y'), ord('e'), terminal.ESCAPE]
    assert terminal.prompt_input('hi') is None

    stdscr.getch.side_effect = [ord('b'), ord('e'), terminal.RETURN]
    assert terminal.prompt_input('hi', key=True) == ord('b')

    stdscr.getch.side_effect = [terminal.ESCAPE, ord('e'), ord('l')]
    assert terminal.prompt_input('hi', key=True) is None


def test_terminal_prompt_y_or_n(terminal, stdscr):

    stdscr.getch.side_effect = [ord('y'), ord('N'), terminal.ESCAPE, ord('a')]
    text = 'hi'.encode('ascii')

    # Press 'y'
    assert terminal.prompt_y_or_n('hi')
    stdscr.subwin.addstr.assert_called_with(0, 0, text)
    assert not curses.flash.called

    # Press 'N'
    assert not terminal.prompt_y_or_n('hi')
    stdscr.subwin.addstr.assert_called_with(0, 0, text)
    assert not curses.flash.called

    # Press Esc
    assert not terminal.prompt_y_or_n('hi')
    stdscr.subwin.addstr.assert_called_with(0, 0, text)
    assert not curses.flash.called

    # Press an invalid key
    assert not terminal.prompt_y_or_n('hi')
    stdscr.subwin.addstr.assert_called_with(0, 0, text)
    assert curses.flash.called


@pytest.mark.parametrize('use_ascii', [True, False])
def test_terminal_open_editor(terminal, use_ascii):

    terminal.config['ascii'] = use_ascii

    comment = COMMENT_EDIT_FILE.format(id=1, content='Comment 1 ❤')
    data = {'filename': None}

    def side_effect(args):
        data['filename'] = args[-1]
        with codecs.open(data['filename'], 'r+', 'utf-8') as fp:
            assert fp.read() == comment
            fp.write('Comment 2 ❤')
        return mock.Mock()

    with mock.patch('subprocess.Popen', autospec=True) as Popen:
        Popen.side_effect = side_effect

        with terminal.open_editor(comment) as reply_text:
            assert reply_text == 'Comment 1 ❤\nComment 2 ❤'
            assert os.path.isfile(data['filename'])
            assert curses.endwin.called
            assert curses.doupdate.called
        assert not os.path.isfile(data['filename'])


def test_terminal_open_editor_error(terminal):

    with mock.patch('subprocess.Popen', autospec=True) as Popen, \
            mock.patch.object(terminal, 'show_notification'):

        # Invalid editor
        Popen.side_effect = OSError
        with terminal.open_editor('hello') as text:
            assert text == 'hello'
        assert 'Could not open' in terminal.show_notification.call_args[0][0]

        data = {'filename': None}

        def side_effect(args):
            data['filename'] = args[-1]
            return mock.Mock()

        # Temporary File Errors don't delete the file
        Popen.side_effect = side_effect
        with terminal.open_editor('test'):
            assert os.path.isfile(data['filename'])
            raise TemporaryFileError()
        assert os.path.isfile(data['filename'])
        os.remove(data['filename'])

        # Other Exceptions don't delete the file *and* are propagated
        Popen.side_effect = side_effect
        with pytest.raises(ValueError):
            with terminal.open_editor('test'):
                assert os.path.isfile(data['filename'])
                raise ValueError()
        assert os.path.isfile(data['filename'])
        os.remove(data['filename'])

        # Gracefully handle the case when we can't remove the file
        with mock.patch.object(os, 'remove'):
            os.remove.side_effect = OSError
            with terminal.open_editor():
                pass
            assert os.remove.called

        assert os.path.isfile(data['filename'])
        os.remove(data['filename'])


def test_terminal_open_link_mailcap(terminal):

    url = 'http://www.test.com'

    class MockMimeParser(object):
        pattern = re.compile('')

    mock_mime_parser = MockMimeParser()

    with mock.patch.object(terminal, 'open_browser'), \
            mock.patch('tuir.terminal.mime_parsers') as mime_parsers:
        mime_parsers.parsers = [mock_mime_parser]

        # Pass through to open_browser if media is disabled
        terminal.config['enable_media'] = False
        terminal.open_link(url)
        assert terminal.open_browser.called
        terminal.open_browser.reset_mock()

        # Invalid content type
        terminal.config['enable_media'] = True
        mock_mime_parser.get_mimetype = lambda url: (url, None)
        terminal.open_link(url)
        assert terminal.open_browser.called
        terminal.open_browser.reset_mock()

        # Text/html defers to open_browser
        mock_mime_parser.get_mimetype = lambda url: (url, 'text/html')
        terminal.open_link(url)
        assert terminal.open_browser.called
        terminal.open_browser.reset_mock()


def test_terminal_open_link_subprocess(terminal):

    url = 'http://www.test.com'
    terminal.config['enable_media'] = True

    with mock.patch('time.sleep'),                            \
            mock.patch('os.system'),                          \
            mock.patch('subprocess.Popen') as Popen,          \
            mock.patch('six.moves.input') as six_input,       \
            mock.patch.object(terminal, 'get_mailcap_entry'):

        six_input.return_values = 'y'

        def reset_mock():
            six_input.reset_mock()
            os.system.reset_mock()
            terminal.stdscr.subwin.addstr.reset_mock()
            Popen.return_value.communicate.return_value = '', 'stderr message'
            Popen.return_value.poll.return_value = 0
            Popen.return_value.wait.return_value = 0

        def get_error():
            # Check if an error message was printed to the terminal
            status = 'Program exited with status'.encode('utf-8')
            return any(status in args[0][2] for args in
                       terminal.stdscr.subwin.addstr.call_args_list)

        # Non-blocking success
        reset_mock()
        entry = ('echo ""', 'echo %s')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert not six_input.called
        assert not get_error()

        # Non-blocking failure
        reset_mock()
        Popen.return_value.poll.return_value = 127
        Popen.return_value.wait.return_value = 127
        entry = ('fake .', 'fake %s')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert not six_input.called
        assert get_error()

        # needsterminal success
        reset_mock()
        entry = ('echo ""', 'echo %s; needsterminal')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert not six_input.called
        assert not get_error()

        # needsterminal failure
        reset_mock()
        Popen.return_value.poll.return_value = 127
        Popen.return_value.wait.return_value = 127
        entry = ('fake .', 'fake %s; needsterminal')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert not six_input.called
        assert get_error()

        # copiousoutput success
        reset_mock()
        entry = ('echo ""', 'echo %s; needsterminal; copiousoutput')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert six_input.called
        assert not get_error()

        # copiousoutput failure
        reset_mock()
        Popen.return_value.poll.return_value = 127
        Popen.return_value.wait.return_value = 127
        entry = ('fake .', 'fake %s; needsterminal; copiousoutput')
        terminal.get_mailcap_entry.return_value = entry
        terminal.open_link(url)
        assert six_input.called
        assert get_error()


def test_terminal_load_mailcaps(stdscr, config):
    with mock.patch('os.path.exists', return_value=True):
        env_mailcaps = os.environ.get('MAILCAPS')
        try:
            os.environ['MAILCAPS'] = 'NON_NULL'

            term = Terminal(stdscr, config=config)
            term._load_mailcaps()

            assert term._mailcap_dict is not None
            assert os.environ['MAILCAPS'] == 'NON_NULL'
        finally:
            if env_mailcaps is None:
                del os.environ['MAILCAPS']
            else:
                os.environ['MAILCAPS'] = env_mailcaps


def test_terminal_open_browser_display(terminal):

    terminal._display = True
    with mock.patch('webbrowser.open_new_tab', autospec=True) as open_new_tab:
        terminal.open_browser('http://www.test.com')

    # open_new_tab() will be executed in the child process so we can't
    # directly check if the was called from here or not.
    # open_new_tab.assert_called_with('http://www.test.com')

    # Shouldn't suspend curses
    assert not curses.endwin.called
    assert not curses.doupdate.called


def test_terminal_open_browser_display_no_response(terminal):

    terminal._display = True
    with mock.patch('tuir.terminal.Process', autospec=True) as Process:
        Process.return_value.is_alive.return_value = 1
        terminal.open_browser('http://www.test.com')
    assert isinstance(terminal.loader.exception, BrowserError)


def test_terminal_open_browser_no_display(terminal):

    terminal._display = False
    with mock.patch('webbrowser.open_new_tab', autospec=True) as open_new_tab:
        terminal.open_browser('http://www.test.com')
    open_new_tab.assert_called_with('http://www.test.com')

    # Should suspend curses to give control of the terminal to the browser
    assert curses.endwin.called
    assert curses.doupdate.called


def test_terminal_open_pager(terminal, stdscr):

    data = "Hello World!  ❤"

    def side_effect(args, stdin=None):
        assert stdin is not None
        raise OSError

    with mock.patch('subprocess.Popen', autospec=True) as Popen, \
            mock.patch.dict('os.environ', {'PAGER': 'fake'}):
        Popen.return_value.stdin = mock.Mock()

        terminal.open_pager(data)
        assert Popen.called
        assert not stdscr.addstr.called

        # Raise an OS error
        Popen.side_effect = side_effect
        terminal.open_pager(data)
        message = 'Could not open pager fake'.encode('ascii')
        print(stdscr.method_calls)
        stdscr.subwin.addstr.assert_called_with(1, 1, message)


def test_terminal_open_urlview(terminal, stdscr):

    data = "Hello World!  ❤"

    def side_effect(args, stdin=None):
        assert stdin is not None
        raise OSError

    with mock.patch('subprocess.Popen') as Popen, \
            mock.patch.dict('os.environ', {'RTV_URLVIEWER': 'fake'}):

        Popen.return_value.poll.return_value = 0
        terminal.open_urlview(data)
        assert Popen.called
        assert not stdscr.addstr.called

        Popen.return_value.poll.return_value = 1
        terminal.open_urlview(data)
        assert stdscr.subwin.addstr.called

        # Raise an OS error
        Popen.side_effect = side_effect
        terminal.open_urlview(data)
        message = 'Failed to open urlview'.encode('utf-8')
        print(stdscr.method_calls)
        stdscr.subwin.addstr.assert_called_with(1, 1, message)


def test_terminal_strip_textpad(terminal):

    assert terminal.strip_textpad(None) is None
    assert terminal.strip_textpad('  foo  ') == '  foo'

    text = 'alpha bravo\ncharlie \ndelta  \n  echo   \n\nfoxtrot\n\n\n'
    assert terminal.strip_textpad(text) == (
        'alpha bravocharlie delta\n  echo\n\nfoxtrot')


def test_terminal_add_space(terminal, stdscr):

    stdscr.x, stdscr.y = 10, 20
    terminal.add_space(stdscr)
    stdscr.addstr.assert_called_with(20, 10, ' ')

    # Not enough room to add a space
    stdscr.reset_mock()
    stdscr.x = 10
    stdscr.ncols = 11
    terminal.add_space(stdscr)
    assert not stdscr.addstr.called


def test_terminal_attr(terminal):

    assert terminal.attr('CursorBlock') == 0
    assert terminal.attr('@CursorBlock') == curses.A_REVERSE
    assert terminal.attr('NeutralVote') == curses.A_BOLD

    with terminal.theme.turn_on_selected():
        assert terminal.attr('CursorBlock') == curses.A_REVERSE
        assert terminal.attr('NeutralVote') == curses.A_BOLD


def test_terminal_check_theme(terminal):

    monochrome = Theme(use_color=False)
    default = Theme()
    color256 = Theme.from_name('molokai')

    curses.has_colors.return_value = False
    assert terminal.check_theme(monochrome)
    assert not terminal.check_theme(default)
    assert not terminal.check_theme(color256)

    curses.has_colors.return_value = True
    curses.COLORS = 0
    assert terminal.check_theme(monochrome)
    assert not terminal.check_theme(default)
    assert not terminal.check_theme(color256)

    curses.COLORS = 8
    assert terminal.check_theme(monochrome)
    assert terminal.check_theme(default)
    assert not terminal.check_theme(color256)

    curses.COLORS = 256
    assert terminal.check_theme(monochrome)
    assert terminal.check_theme(default)
    assert terminal.check_theme(color256)

    curses.COLOR_PAIRS = 8
    assert terminal.check_theme(monochrome)
    assert terminal.check_theme(default)
    assert not terminal.check_theme(color256)


def test_terminal_set_theme(terminal, stdscr):

    # Default with color enabled
    stdscr.reset_mock()
    terminal.set_theme()
    assert terminal.theme.use_color
    assert terminal.theme.display_string == 'default (built-in)'
    stdscr.bkgd.assert_called_once_with(' ', 0)

    # When the user passes in the --monochrome flag
    terminal.theme = None
    terminal.set_theme(Theme(use_color=False))
    assert not terminal.theme.use_color
    assert terminal.theme.display_string == 'monochrome (built-in)'

    # When the terminal doesn't support colors
    curses.COLORS = 0
    terminal.theme = None
    terminal.set_theme()
    assert terminal.theme.display_string == 'monochrome (built-in)'

    # When the terminal doesn't support 256 colors so it falls back to the
    # built-in default theme
    curses.COLORS = 8
    terminal.theme = None
    terminal.set_theme(Theme.from_name('molokai'))
    assert terminal.theme.display_string == 'default (built-in)'

    # When the terminal does support the 256 color theme
    curses.COLORS = 256
    terminal.theme = None
    terminal.set_theme(Theme.from_name('molokai'))
    assert terminal.theme.display_string == 'molokai (preset)'


def test_terminal_set_theme_no_colors(terminal, stdscr):

    # Monochrome should be forced if the terminal doesn't support color
    with mock.patch('curses.has_colors') as has_colors:
        has_colors.return_value = False

        terminal.set_theme()
        assert not terminal.theme.use_color

        terminal.set_theme(Theme(use_color=True))
        assert not terminal.theme.use_color


def test_terminal_strip_instructions(terminal):

    # These templates only contain instructions, so they should be empty
    assert terminal.strip_instructions(SUBMISSION_FILE) == ''
    assert terminal.strip_instructions(REPLY_FILE) == ''
    assert terminal.strip_instructions(MESSAGE_FILE) == ''

    # These templates should strip everything but the {content} tag,
    # which will be replaced with the submission/content to be edited
    assert terminal.strip_instructions(SUBMISSION_EDIT_FILE) == '{content}'
    assert terminal.strip_instructions(COMMENT_EDIT_FILE) == '{content}'

    # Comments without the INSTRUCTIONS marker shouldn't be stripped
    text = '<!-- A normal HTML comment -->'
    assert terminal.strip_instructions(text) == text

    # Hashes shouldn't be stripped anymore
    text = '# This is no longer interpreted as a comment!'
    assert terminal.strip_instructions(text) == text

    # Nested HTML comment tags shouldn't be shown
    text = '<!--{0} <!-- A nested HTML comment --> {0}-->'.format(TOKEN)
    assert terminal.strip_instructions(text) == ''

    # Another edge case
    text = '<!--{0} instructions {0}--><!-- An HTML comment -->'.format(TOKEN)
    assert terminal.strip_instructions(text) == '<!-- An HTML comment -->'

    # Whitespace at the start of the first line should be preserved
    text = '    code\n    block\n'
    assert terminal.strip_instructions(text) == '    code\n    block'

    # But blank lines should be stripped
    text = '\n    \n    code\n    block\n'
    assert terminal.strip_instructions(text) == '    code\n    block'

    # However, blank lines in the middle of comment should not be stripped
    text = 'paragraph 1\n\nparagraph 2\n'
    assert terminal.strip_instructions(text) == 'paragraph 1\n\nparagraph 2'


def test_terminal_get_link_pages(terminal):

    # When there are no links passed in, there should be no pages generated
    assert terminal.get_link_pages([]) == []

    # A single link should generate a single page
    assert terminal.get_link_pages([1]) == [[1]]

    # Up to 10 links should fit on a single page
    link_pages = terminal.get_link_pages([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    assert link_pages == [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]

    # When the 11th link is added a new page is created
    link_pages = terminal.get_link_pages([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    assert link_pages == [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [10]]


def test_terminal_get_link_page_text(terminal):

    link_page = [
        {'href': 'https://www.reddit.com', 'text': 'Reddit Homepage'},
        {'href': 'https://www.duckduckgo.com', 'text': 'Search Engine'},
        {
            'href': 'https://gitlab.com/ajak/tuir',
            'text': 'This project\'s homepage'
        }
    ]

    text = terminal.get_link_page_text(link_page)
    assert text == dedent("""\
    [0] [Reddit Homepage](https://www.reddit.com)
    [1] [Search Engine](https://www.duckduckgo.com)
    [2] [This project's home…](https://gitlab.com/ajak/tuir)
    """)


def test_terminal_prompt_user_to_select_link(terminal, stdscr):

    # Select the 2nd link on the 2nd page
    links = [
        {'href': 'www.website-0.com', 'text': 'Website 0'},
        {'href': 'www.website-1.com', 'text': 'Website 1'},
        {'href': 'www.website-2.com', 'text': 'Website 2'},
        {'href': 'www.website-3.com', 'text': 'Website 3'},
        {'href': 'www.website-4.com', 'text': 'Website 4'},
        {'href': 'www.website-5.com', 'text': 'Website 5'},
        {'href': 'www.website-6.com', 'text': 'Website 6'},
        {'href': 'www.website-7.com', 'text': 'Website 7'},
        {'href': 'www.website-8.com', 'text': 'Website 8'},
        {'href': 'www.website-9.com', 'text': 'Website 9'},
        {'href': 'www.website-10.com', 'text': 'Website 10'},
        {'href': 'www.website-11.com', 'text': 'Website 11'},
        {'href': 'www.website-12.com', 'text': 'Website 12'},
    ]
    stdscr.getch.side_effect = [ord('j'), ord('2')]
    href = terminal.prompt_user_to_select_link(links)
    assert href == 'www.website-12.com'

    # Select a link that doesn't exist
    links = [
        {'href': 'www.website-0.com', 'text': 'Website 0'},
        {'href': 'www.website-1.com', 'text': 'Website 1'},
        {'href': 'www.website-2.com', 'text': 'Website 2'},
    ]
    stdscr.getch.side_effect = [ord('3')]
    assert terminal.prompt_user_to_select_link(links) is None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import curses
from collections import OrderedDict
import os.path
import sys

import six
import pytest

from tuir import __version__
from tuir.subreddit_page import SubredditPage
from tuir.packages.praw.errors import NotFound, HTTPException
from requests.exceptions import ReadTimeout

try:
    from unittest import mock
except ImportError:
    import mock


PROMPTS = OrderedDict([
    ('prompt_1', 'comments/571dw3'),
    ('prompt_2', '///comments/571dw3'),
    ('prompt_3', '/comments/571dw3'),
    ('prompt_4', '/r/pics/comments/571dw3/'),
    ('prompt_5', 'https://www.reddit.com/r/pics/comments/571dw3/at_disneyland'),
])


def test_subreddit_page_construct_default(reddit, terminal, config, oauth):
    config['look_and_feel'] = 'default'

    window = terminal.stdscr.subwin

    with terminal.loader():
        page = SubredditPage(reddit, terminal, config, oauth, '/r/python')
    assert terminal.loader.exception is None
    page.draw()

    # Title
    title = '/r/python'.encode('utf-8')
    window.addstr.assert_any_call(0, 0, title)

    # Banner
    menu = '[1]hot     [2]top     [3]rising     [4]new     [5]controversial     [6]gilded'.encode('utf-8')
    window.addstr.assert_any_call(0, 0, menu)

    # Submission
    text = page.content.get(0)['split_title'][0].encode('utf-8')
    window.subwin.addstr.assert_any_call(0, 1, text, 2097152)

    # Cursor should have been drawn
    window.subwin.addch.assert_any_call(0, 0, ' ', curses.A_REVERSE)

    # Reload with a smaller terminal window
    terminal.stdscr.ncols = 20
    terminal.stdscr.nlines = 10
    with terminal.loader():
        page = SubredditPage(reddit, terminal, config, oauth, '/r/python')
    assert terminal.loader.exception is None
    page.draw()


def test_subreddit_refresh(subreddit_page, terminal):

    # Refresh the page with default values
    subreddit_page.controller.trigger('r')
    assert subreddit_page.content.order is None
    assert subreddit_page.content.name == '/r/python'
    assert terminal.loader.exception is None

    # Refresh with the order in the name
    subreddit_page.refresh_content(order='ignore', name='/r/front/hot')
    assert subreddit_page.content.order == 'hot'
    assert subreddit_page.content.name == '/r/front'
    assert terminal.loader.exception is None


def test_subreddit_reload_page(subreddit_page, terminal, reddit):

    cache = reddit.handler.cache
    assert len(cache) == 1

    # A plain refresh_content() will use whatever is in the praw cache
    # instead of making a new request to reddit
    list(cache.values())[0].status_code = 503
    subreddit_page.refresh_content()
    assert isinstance(terminal.loader.exception, HTTPException)

    cache = reddit.handler.cache
    assert len(cache) == 1

    # But if we manually trigger a page refresh, it should clear the cache
    # and reload the page instead of returning the cached 503 response
    list(cache.values())[0].status_code = 503
    subreddit_page.controller.trigger('r')
    assert terminal.loader.exception is None


def test_subreddit_title(subreddit_page, terminal, capsys):
    subreddit_page.content.name = 'hello ❤'

    with mock.patch.dict('os.environ', {'DISPLAY': ':1'}):
        terminal.config['ascii'] = True
        subreddit_page.draw()
        out, _ = capsys.readouterr()
        assert isinstance(out, six.text_type)
        assert out == '\x1b]2;hello ? - tuir {}\x07'.format(__version__)

        terminal.config['ascii'] = False
        subreddit_page.draw()
        out, _ = capsys.readouterr()
        assert isinstance(out, six.text_type)
        assert out == '\x1b]2;hello ❤ - tuir {}\x07'.format(__version__)

    with mock.patch.dict('os.environ', {'DISPLAY': ''}):
        subreddit_page.draw()
        out, _ = capsys.readouterr()
        assert not out

    with mock.patch.dict('os.environ', {'INSIDE_EMACS': '25.3.1,term:0.96'}):
        subreddit_page.draw()
        out, _ = capsys.readouterr()
        assert not out


def test_subreddit_search(subreddit_page, terminal):
    window = terminal.stdscr.subwin

    # Search the current subreddit
    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = 'search term'
        subreddit_page.controller.trigger('f')
        assert subreddit_page.content.name == '/r/python'
        assert terminal.prompt_input.called
        assert not terminal.loader.exception

        # The page title should display the query
        subreddit_page.draw()
        title = 'Searching /r/python: search term'.encode('utf-8')
        window.addstr.assert_any_call(0, 0, title)

    # Ordering the results should preserve the query
    window.addstr.reset_mock()
    subreddit_page.refresh_content(order='hot')
    subreddit_page.refresh_content(order='top-all')
    subreddit_page.refresh_content(order='new')
    assert subreddit_page.content.name == '/r/python'
    assert subreddit_page.content.query == 'search term'
    assert not terminal.loader.exception

    # Searching with an empty query shouldn't crash
    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = None
        subreddit_page.controller.trigger('f')
        assert not terminal.loader.exception

    # Changing to a new subreddit should clear the query
    window.addstr.reset_mock()
    subreddit_page.refresh_content(name='/r/learnpython')
    assert subreddit_page.content.query is None


def test_subreddit_prompt(subreddit_page, terminal):

    # Prompt for a different subreddit
    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = 'front/top'
        subreddit_page.controller.trigger('/')

        subreddit_page.handle_selected_page()
        assert not subreddit_page.active
        assert subreddit_page.selected_page
        assert subreddit_page.selected_page.content.name == '/r/front'
        assert subreddit_page.selected_page.content.order == 'top'


@pytest.mark.parametrize('prompt', PROMPTS.values(), ids=list(PROMPTS))
def test_subreddit_prompt_submission(subreddit_page, terminal, prompt):

    url = 'https://www.reddit.com/comments/571dw3'

    with mock.patch.object(subreddit_page, 'open_submission_page'), \
            mock.patch.object(terminal, 'prompt_input'):

        terminal.prompt_input.return_value = prompt
        subreddit_page.open_submission_page.return_value = 'MockPage'
        subreddit_page.controller.trigger('/')

        subreddit_page.open_submission_page.assert_called_with(url)
        assert not terminal.loader.exception
        assert subreddit_page.selected_page == 'MockPage'


def test_subreddit_prompt_submission_invalid(subreddit_page, terminal):

    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = 'comments/571dw3fakeid'
        subreddit_page.controller.trigger('/')
        assert isinstance(terminal.loader.exception, NotFound)


def test_subreddit_order(subreddit_page):

    # /r/python doesn't always have rising submissions, so use a larger sub
    subreddit_page.refresh_content(name='all')

    subreddit_page.content.query = ''
    subreddit_page.controller.trigger('1')
    assert subreddit_page.content.order == 'hot'
    subreddit_page.controller.trigger('3')
    assert subreddit_page.content.order == 'rising'
    subreddit_page.controller.trigger('4')
    assert subreddit_page.content.order == 'new'
    subreddit_page.controller.trigger('6')
    assert subreddit_page.content.order == 'gilded'

    subreddit_page.content.query = 'search text'
    subreddit_page.controller.trigger('1')
    assert subreddit_page.content.order == 'relevance'
    subreddit_page.controller.trigger('4')
    assert subreddit_page.content.order == 'new'

    # Shouldn't be able to sort queries by gilded
    subreddit_page.controller.trigger('6')
    assert curses.flash.called
    assert subreddit_page.content.order == 'new'


def test_subreddit_order_top(subreddit_page, terminal):

    # Sort by top
    with mock.patch.object(terminal, 'show_notification'):
        # Invalid selection
        terminal.show_notification.return_value = ord('x')
        subreddit_page.controller.trigger('2')
        terminal.show_notification.assert_called_with('Invalid option')
        assert subreddit_page.content.order is None

        # Valid selection - sort by week
        terminal.show_notification.reset_mock()
        terminal.show_notification.return_value = ord('3')
        subreddit_page.controller.trigger('2')
        assert subreddit_page.content.order == 'top-week'


def test_subreddit_order_controversial(subreddit_page, terminal):

    # Sort by controversial
    with mock.patch.object(terminal, 'show_notification'):
        # Invalid selection
        terminal.show_notification.return_value = ord('x')
        subreddit_page.controller.trigger('5')
        terminal.show_notification.assert_called_with('Invalid option')
        assert subreddit_page.content.order is None

        # Valid selection - sort by default
        terminal.show_notification.reset_mock()
        terminal.show_notification.return_value = ord('\n')
        subreddit_page.controller.trigger('5')
        assert subreddit_page.content.order == 'controversial'


def test_subreddit_order_search(subreddit_page, terminal):

    # Search the current subreddit
    with mock.patch.object(terminal, 'prompt_input'):
        terminal.prompt_input.return_value = 'search term'
        subreddit_page.controller.trigger('f')
        assert subreddit_page.content.name == '/r/python'
        assert terminal.prompt_input.called
        assert not terminal.loader.exception

    # Sort by relevance
    subreddit_page.controller.trigger('1')
    assert subreddit_page.content.order == 'relevance'

    # Sort by top
    with mock.patch.object(terminal, 'show_notification'):
        terminal.show_notification.reset_mock()
        terminal.show_notification.return_value = ord('6')
        subreddit_page.controller.trigger('2')
        assert subreddit_page.content.order == 'top-all'

    # Sort by comments
    with mock.patch.object(terminal, 'show_notification'):
        terminal.show_notification.reset_mock()
        terminal.show_notification.return_value = ord('6')
        subreddit_page.controller.trigger('3')
        assert subreddit_page.content.order == 'comments-all'

    # Sort by new
    subreddit_page.controller.trigger('4')
    assert subreddit_page.content.order == 'new'


def test_subreddit_open(subreddit_page, terminal, config):

    # Open the selected submission
    data = subreddit_page.content.get(subreddit_page.nav.absolute_index)
    with mock.patch.object(config.history, 'add'):
        data['url_type'] = 'selfpost'
        subreddit_page.controller.trigger('l')
        assert not terminal.loader.exception
        assert subreddit_page.selected_page
        assert subreddit_page.active
        config.history.add.assert_called_with(data['url_full'])

    # Open the selected link externally
    data = subreddit_page.content.get(subreddit_page.nav.absolute_index)
    with mock.patch.object(terminal, 'open_link'), \
            mock.patch.object(config.history, 'add'):
        data['url_type'] = 'external'
        subreddit_page.controller.trigger('o')
        assert terminal.open_link.called
        config.history.add.assert_called_with(data['url_full'])

    # Open the selected link within tuir
    data = subreddit_page.content.get(subreddit_page.nav.absolute_index)
    with mock.patch.object(subreddit_page, 'open_submission'), \
            mock.patch.object(config.history, 'add'):
        data['url_type'] = 'selfpost'
        subreddit_page.controller.trigger('o')
        assert subreddit_page.open_submission.called


def test_subreddit_open_xpost(subreddit_page, config):

    data = subreddit_page.content.get(subreddit_page.nav.absolute_index)

    # Open an x-post subreddit, see /r/TinySubredditoftheDay for an example
    with mock.patch.object(subreddit_page, 'refresh_content'):
        data['url_type'] = 'x-post subreddit'
        data['xpost_subreddit'] = 'goodbye'
        subreddit_page.controller.trigger('o')
        subreddit_page.refresh_content.assert_called_with(
            name='goodbye', order='ignore')

    # Open an x-post submission, see /r/bestof for an example
    with mock.patch.object(subreddit_page, 'open_submission'):
        data['url_type'] = 'x-post submission'
        data['url_full'] = 'www.test.com'
        subreddit_page.controller.trigger('o')
        subreddit_page.open_submission.assert_called_with(url='www.test.com')


def test_subreddit_unauthenticated(subreddit_page, terminal):

    # Unauthenticated commands
    methods = [
        'a',  # Upvote
        'z',  # Downvote
        'c',  # Post
        'e',  # Edit
        'd',  # Delete
        's',  # Subscriptions
    ]
    for ch in methods:
        subreddit_page.controller.trigger(ch)
        text = 'Not logged in'.encode('utf-8')
        terminal.stdscr.subwin.addstr.assert_called_with(1, 1, text)


def test_subreddit_post(subreddit_page, terminal, reddit, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # Post a submission to an invalid subreddit
    subreddit_page.refresh_content(name='front')
    subreddit_page.controller.trigger('c')
    text = "Can't post to /r/front".encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_called_with(1, 1, text)

    # Post a submission with a title but with no body
    subreddit_page.refresh_content(name='python')
    with mock.patch.object(terminal, 'open_editor'):
        terminal.open_editor.return_value.__enter__.return_value = 'title'
        subreddit_page.controller.trigger('c')
        text = 'Missing body'.encode('utf-8')
        terminal.stdscr.subwin.addstr.assert_called_with(1, 1, text)

    # Post a fake submission
    url = 'https://www.reddit.com/r/Python/comments/2xmo63/'
    submission = reddit.get_submission(url)
    with mock.patch.object(terminal, 'open_editor'),  \
            mock.patch.object(reddit, 'submit'),      \
            mock.patch('time.sleep'):
        terminal.open_editor.return_value.__enter__.return_value = 'test\ncont'
        reddit.submit.return_value = submission
        subreddit_page.controller.trigger('c')
        assert reddit.submit.called
        assert subreddit_page.selected_page.content._submission == submission
        assert subreddit_page.active


def test_subreddit_open_subscriptions(subreddit_page, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # Open subscriptions
    subreddit_page.controller.trigger('s')
    assert subreddit_page.selected_page
    assert subreddit_page.active

    with mock.patch('tuir.page.Page.loop') as loop:
        subreddit_page.handle_selected_page()
        assert loop.called


def test_subreddit_get_inbox_timeout(subreddit_page, refresh_token, terminal, vcr):
    if vcr.record_mode == 'none':
        pytest.skip('Unable to test ReadTimeout exceptions using a cassette')

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    subreddit_page.reddit.config.timeout = 0.00000001
    subreddit_page.controller.trigger('i')
    text = 'HTTP request timed out'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_called_with(1, 1, text)
    assert isinstance(terminal.loader.exception, ReadTimeout)


def test_subreddit_open_multireddits(subreddit_page, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # Open multireddits
    subreddit_page.controller.trigger('S')
    assert subreddit_page.selected_page
    assert subreddit_page.active

    with mock.patch('tuir.page.Page.loop') as loop:
        subreddit_page.handle_selected_page()
        assert loop.called


def test_subreddit_private_user_pages(subreddit_page, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    subreddit_page.refresh_content(name='/u/me/saved')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/hidden')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/upvoted')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/downvoted')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/overview')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/submitted')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/me/comments')
    subreddit_page.draw()


def test_subreddit_user_pages(subreddit_page, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # Pick a user that has a lot of recent comments, so we can make sure that
    # SavedComment objects have all of the properties necessary to be drawn
    # on the submission page.

    # Should default to the overview page
    subreddit_page.refresh_content(name='/u/spez')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/spez/overview')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/spez/submitted')
    subreddit_page.draw()

    subreddit_page.refresh_content(name='/u/spez/comments')
    subreddit_page.draw()


def test_subreddit_draw_header(subreddit_page, refresh_token, terminal):

    # /r/front alias should be renamed in the header
    subreddit_page.refresh_content(name='/r/front')
    subreddit_page.draw()
    text = 'Front Page'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    subreddit_page.refresh_content(name='/r/front/new')
    subreddit_page.draw()
    text = 'Front Page'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    # Log in to check the user submissions page
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # /u/me alias should be renamed in the header
    subreddit_page.refresh_content(name='/u/me')
    subreddit_page.draw()
    text = 'My Overview'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    subreddit_page.refresh_content(name='/u/me/new')
    subreddit_page.draw()
    text = 'My Overview'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    # /u/saved alias should be renamed in the header
    subreddit_page.refresh_content(name='/u/me/saved')
    subreddit_page.draw()
    text = 'My Saved Content'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    # /u/upvoted alias should be renamed in the header
    subreddit_page.refresh_content(name='/u/me/upvoted')
    subreddit_page.draw()
    text = 'My Upvoted Content'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    # /u/downvoted alias should be renamed in the header
    subreddit_page.refresh_content(name='/u/me/downvoted')
    subreddit_page.draw()
    text = 'My Downvoted Content'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)

    # /u/hidden alias should be renamed in the header
    subreddit_page.refresh_content(name='/u/me/hidden')
    subreddit_page.draw()
    text = 'My Hidden Content'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(0, 0, text)


def test_subreddit_frontpage_toggle(subreddit_page, terminal):
    with mock.patch.object(terminal, 'prompt_input'):

        terminal.prompt_input.return_value = 'aww'
        subreddit_page.controller.trigger('/')
        subreddit_page.handle_selected_page()

        new_page = subreddit_page.selected_page
        assert new_page is not None
        assert new_page.content.name == '/r/aww'

        new_page.controller.trigger('p')
        assert new_page.toggled_subreddit == '/r/aww'
        assert new_page.content.name == '/r/front'


def test_subreddit_hide_submission(subreddit_page, refresh_token):

    # Log in
    subreddit_page.config.refresh_token = refresh_token
    subreddit_page.oauth.authorize()

    # The api won't return hidden posts in the submission listing, so the
    # first post should always have hidden set to false
    data = subreddit_page.get_selected_item()
    assert data['hidden'] is False

    # Hide the first submission by pressing the space key
    subreddit_page.controller.trigger(0x20)
    assert subreddit_page.term.loader.exception is None
    data = subreddit_page.get_selected_item()
    assert data['hidden'] is True

    # Make sure that the status was actually updated on the server side
    data['object'].refresh()
    assert data['object'].hidden is True

    # Now undo the hide by pressing space again
    subreddit_page.controller.trigger(0x20)
    assert subreddit_page.term.loader.exception is None
    data = subreddit_page.get_selected_item()
    assert data['hidden'] is False

    # Make sure that the status was actually updated on the server side
    data['object'].refresh()
    assert data['object'].hidden is False


def test_subreddit_handle_selected_page(subreddit_page, subscription_page):

    # Method should be a no-op if selected_page is unset
    subreddit_page.active = True
    subreddit_page.handle_selected_page()
    assert subreddit_page.selected_page is None
    assert subreddit_page.active

    # Open the subscription page and select a subreddit from the list of
    # subscriptions
    with mock.patch.object(subscription_page, 'loop', return_value=subreddit_page):
        subreddit_page.selected_page = subscription_page
        subreddit_page.handle_selected_page()
        assert subreddit_page.selected_page == subreddit_page
        assert subreddit_page.active

    # Now when handle_select_page() is called again, the current subreddit
    # should be closed so the selected page can be opened
    subreddit_page.handle_selected_page()
    assert subreddit_page.selected_page == subreddit_page
    assert not subreddit_page.active

    subreddit_page.selected_page.name = 'This should raise a RuntimeError'
    with pytest.raises(RuntimeError):
        subreddit_page.handle_selected_page()


def test_subreddit_page_loop_pre_select(subreddit_page, submission_page):

    # Set the selected_page before entering the loop(). This will cause the
    # selected page to immediately open. If the selected page returns a
    # different subreddit page (e.g. the user enters a subreddit into the
    # prompt before they hit the `h` key), the initial loop should be closed
    # immediately
    subreddit_page.selected_page = submission_page
    with mock.patch.object(submission_page, 'loop', return_value=subreddit_page):
        selected_page = subreddit_page.loop()

        assert not subreddit_page.active
        assert selected_page == subreddit_page


def test_subreddit_page_loop(subreddit_page, stdscr, terminal):

    stdscr.getch.return_value = ord('/')

    with mock.patch.object(terminal, 'prompt_input', return_value='all'):
        new_page = subreddit_page.loop()
        assert new_page.content.name == '/r/all'


def test_subreddit_page__submission_attr(config, terminal, subreddit_page):
    data = {}
    data['url_full'] = 'www.test.com'

    config.history = []
    terminal.attr = mock.Mock()

    subreddit_page._submission_attr(data)
    terminal.attr.assert_called_with('SubmissionTitle')

    config.history.append(data['url_full'])

    subreddit_page._submission_attr(data)
    terminal.attr.assert_called_with('SubmissionTitleSeen')


def test_subreddit_page__url_str(config, terminal, subreddit_page):
    data = {
            'url_type': 'selfpost',
            'url': 'self.AskReddit',
            'url_full': 'https://www.reddit.com/r/AskReddit/comments/99eh6b/without_saying_what_the_category_is_what_are_your/'
    }

    assert subreddit_page._url_str(data) == 'self.AskReddit'

    data['url_type'] = 'x-post subreddit'

    assert subreddit_page._url_str(data) == 'self.AskReddit'

    data['url_type'] = 'external'

    assert subreddit_page._url_str(data) == 'www.reddit.com'


def test_subreddit_page__url_attr(config, terminal, subreddit_page):
    data = {}
    data['url_full'] = 'www.test.com'

    config.history = []
    terminal.attr = mock.Mock()

    subreddit_page._url_attr(data)
    terminal.attr.assert_called_with('Link')

    config.history.append(data['url_full'])

    subreddit_page._url_attr(data)
    terminal.attr.assert_called_with('LinkSeen')


def test_subreddit_page__gold_str(config, terminal, subreddit_page):
    data = {}
    data['gold'] = 0

    goldstr = subreddit_page._gold_str(data)
    assert goldstr == ''

    data['gold'] = 1
    goldstr = subreddit_page._gold_str(data)
    assert goldstr == terminal.gilded

    data['gold'] = 2
    goldstr = subreddit_page._gold_str(data)
    assert goldstr == terminal.gilded + 'x2'


def test_subreddit_page__draw_item_format(terminal, subreddit_page):
    data = {
            'index': '1',
            'title': 'Without saying what the category is, what are your top five?',
            'score': '144655',
            'likes': True,
            'comments': '26584',
            'created': '10month',
            'edited': '(edit 5month)',
            'author': 'reddit_user',
            'subreddit': 'AskReddit',
            'url': 'self.AskReddit',
            'url_type': 'selfpost',
            'url_full': 'https://www.reddit.com/r/AskReddit/comments/99eh6b/without_saying_what_the_category_is_what_are_your/',
            'gold': 1,
            'saved': True,
            'hidden': True,
            'stickied': True,
            'nsfw': True,
            'flair': 'Serious Replies Only'
    }

    expected_str = {
            '%i': '1',
            '%t': 'Without saying what the category is, what are your top five?',
            '%s': '144655',
            '%v': '▲',
            '%c': '26584',
            '%r': '10month',
            '%e': '(edit 5month)',
            '%a': 'reddit_user',
            '%S': '/r/AskReddit',
            '%u': 'self.AskReddit',
            '%U': 'self.AskReddit',
            '%A': '[saved]',
            '%h': '[hidden]',
            '%T': '[stickied]',
            '%g': '✪',
            '%n': 'NSFW',
            '%f': 'Serious Replies Only',
    }

    expected_attr = {
            '%i': terminal.attr('SubmissionTitle'),
            '%t': terminal.attr('SubmissionTitle'),
            '%s': terminal.attr('Score'),
            '%v': terminal.attr('Upvote'),
            '%c': terminal.attr('CommentCount'),
            '%r': terminal.attr('Created'),
            '%e': terminal.attr('Created'),
            '%a': terminal.attr('SubmissionAuthor'),
            '%S': terminal.attr('SubmissionSubreddit'),
            '%u': terminal.attr('Link'),
            '%U': terminal.attr('Link'),
            '%A': terminal.attr('Saved'),
            '%h': terminal.attr('Hidden'),
            '%T': terminal.attr('Stickied'),
            '%g': terminal.attr('Gold'),
            '%n': terminal.attr('NSFW'),
            '%f': terminal.attr('SubmissionFlair')
    }

    win = terminal.stdscr.subwin

    with mock.patch.object(terminal, 'add_line'):
        # Loop through each individual format specifier and check for the desired behavior.
        for fmt in filter(None, '%i %t %s %v %c %r %e %a %S %u %U %A %h %T %g %n %f %F'.split()):
                subreddit_page.config['subreddit_format'] = fmt
                subreddit_page.FORMAT_LIST = subreddit_page._create_format_list()
                subreddit_page._draw_item_format(win, data, 0, 0)

                try:
                    if fmt == '%F':

                        # Make sure each of the things that %F should print are printed
                        # flair
                        terminal.add_line.assert_any_call(win, expected_str['%f'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%f'])

                        # saved
                        terminal.add_line.assert_any_call(win, expected_str['%A'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%A'])

                        # hidden
                        terminal.add_line.assert_any_call(win, expected_str['%h'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%h'])

                        # stickied
                        terminal.add_line.assert_any_call(win, expected_str['%T'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%T'])

                        # gold
                        terminal.add_line.assert_any_call(win, expected_str['%g'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%g'])

                        # nsfw
                        terminal.add_line.assert_any_call(win, expected_str['%n'],
                                                          row=mock.ANY, col=mock.ANY,
                                                          attr=expected_attr['%n'])
                    else:
                        terminal.add_line.assert_called_with(win, expected_str[fmt],
                                                             row=0, col=1,
                                                             attr=expected_attr[fmt])
                except:
                    sys.stderr.write("Failed on {0}\n".format(fmt))
                    raise

        terminal.add_line.reset_mock()

        # Ensure spaces aren't printed consecutively if data is absent
        data['gold'] = 0
        subreddit_page.config['subreddit_format'] = ' %g '
        subreddit_page.FORMAT_LIST = subreddit_page._create_format_list()
        subreddit_page._draw_item_format(win, data, 0, 0)

        assert terminal.add_line.call_count == 1

        terminal.add_line.reset_mock()

        # Test for correct handling of separators
        subreddit_page.config['subreddit_format'] = ' | '
        subreddit_page.FORMAT_LIST = subreddit_page._create_format_list()
        subreddit_page._draw_item_format(win, data, 0, 0)

        # Should be called thrice - ' ', '|', ' '
        assert terminal.add_line.call_count == 3


class MockPopen():
    def __init__(*args, **kwargs): pass
    def communicate(self): return '', ''
    def poll(self): pass


class MockProcess():
    def __init__(self, target, args=None):
        self._target = target
        self._args = args or []
    def start(self):
        self._target(*self._args)
    def join(self, _=None): pass
    def is_alive(self): pass
    def terminate(self): pass


def test_subreddit_open_in_browser(subreddit_page, terminal):
    subreddit_page.refresh_content(order='ignore', name='/r/front/hot')
    terminal.config.MAILCAP = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'tuir', 'templates', 'mailcap')
    with mock.patch('os.system'), \
            mock.patch('tuir.terminal.Process', side_effect=MockProcess), \
            mock.patch('webbrowser.open_new_tab') as new_tab:
        for i in range(8):
            subreddit_page.controller.trigger('o')
            with mock.patch.object(subreddit_page, 'clear_input_queue'):
                subreddit_page.controller.trigger('j')
                subreddit_page.controller.trigger('j')
                subreddit_page.controller.trigger('j')
        assert [list(i.args) for i in new_tab.mock_calls] == [
            ['https://i.redd.it/czlc1ff2tqlz.jpg'],
            ['https://i.imgur.com/ok9H05N.gifv'],
            ['https://twitter.com/DOOM/status/908094489276928001'],
            ['https://i.imgur.com/qrWbOue.gifv'],
            ['https://i.redd.it/3lld01w00qlz.gif'],
            ['https://i.imgur.com/Kt6HUl8.jpg'],
            ['https://i.redd.it/2rbvjstdfplz.jpg'],
            ['https://twitter.com/elonmusk/status/908108029777686528'],
        ]


def test_subreddit_open_with_mailcap(subreddit_page, terminal):
    subreddit_page.refresh_content(order='ignore', name='/r/front/hot')
    terminal.config['enable_media'] = True
    terminal.config.MAILCAP = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'tuir', 'templates', 'mailcap')
    with mock.patch('os.system'), \
            mock.patch('tuir.terminal.Process', side_effect=MockProcess), \
            mock.patch('subprocess.Popen', side_effect=MockPopen) as popen, \
            mock.patch('webbrowser.open_new_tab') as new_tab, \
            mock.patch('six.moves.input'):
        for i in range(8):
            subreddit_page.controller.trigger('o')
            with mock.patch.object(subreddit_page, 'clear_input_queue'):
                subreddit_page.controller.trigger('j')
                subreddit_page.controller.trigger('j')
                subreddit_page.controller.trigger('j')
        assert [list(i.args) for i in new_tab.mock_calls] == [
            ['https://twitter.com/DOOM/status/908094489276928001'],
            ['https://twitter.com/elonmusk/status/908108029777686528'],
        ]
        assert [list(i.args) for i in popen.mock_calls] == [
            [["curl -s 'https://i.redd.it/czlc1ff2tqlz.jpg' | convert -resize 80x80 - jpg:/tmp/tuir.jpg && img2xterm /tmp/tuir.jpg"]],
            [["mpv -vo drm -quiet 'https://i.imgur.com/ok9H05N.mp4'"]],
            [["mpv -vo drm -quiet 'https://i.imgur.com/qrWbOue.mp4'"]],
            [["curl -s 'https://i.redd.it/3lld01w00qlz.gif' | convert -resize 80x80 - jpg:/tmp/tuir.jpg && img2xterm /tmp/tuir.jpg"]],
            [["curl -s 'https://i.imgur.com/Kt6HUl8.jpg' | convert -resize 80x80 - jpg:/tmp/tuir.jpg && img2xterm /tmp/tuir.jpg"]],
            [["curl -s 'https://i.redd.it/2rbvjstdfplz.jpg' | convert -resize 80x80 - jpg:/tmp/tuir.jpg && img2xterm /tmp/tuir.jpg"]],
        ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import requests

from tuir.oauth import OAuthHelper, OAuthHandler
from tuir.exceptions import InvalidRefreshToken
from tuir.packages.praw.errors import OAuthException


try:
    from unittest import mock
except ImportError:
    import mock


def test_oauth_handler_not_found(oauth_server):

    url = oauth_server.url + 'favicon.ico'
    resp = requests.get(url)
    assert resp.status_code == 404


def test_oauth_handler_no_callback(oauth_server):

    resp = requests.get(oauth_server.url)
    assert resp.status_code == 200
    assert 'Wait...' in resp.text
    assert OAuthHandler.params['error'] is None


def test_oauth_handler_access_denied(oauth_server):

    url = oauth_server.url + '?error=access_denied'
    resp = requests.get(url)
    assert resp.status_code == 200
    assert OAuthHandler.params['error'] == 'access_denied'
    assert 'denied access' in resp.text


def test_oauth_handler_error(oauth_server):

    url = oauth_server.url + '?error=fake'
    resp = requests.get(url)
    assert resp.status_code == 200
    assert OAuthHandler.params['error'] == 'fake'
    assert 'fake' in resp.text


def test_oauth_handler_success(oauth_server):

    url = oauth_server.url + '?state=fake_state&code=fake_code'
    resp = requests.get(url)
    assert resp.status_code == 200
    assert OAuthHandler.params['error'] is None
    assert OAuthHandler.params['state'] == 'fake_state'
    assert OAuthHandler.params['code'] == 'fake_code'
    assert 'Access Granted' in resp.text


def test_oauth_terminal_non_mobile_authorize(reddit, terminal, config):

    # Should direct to the desktop version if using a graphical browser
    reddit.config.API_PATHS['authorize'] = 'api/v1/authorize/'
    terminal._display = True
    oauth = OAuthHelper(reddit, terminal, config)
    assert '.compact' not in oauth.reddit.config.API_PATHS['authorize']


def test_oauth_terminal_mobile_authorize(reddit, terminal, config):

    # Should direct to the mobile version if using a terminal browser
    reddit.config.API_PATHS['authorize'] = 'api/v1/authorize/'
    terminal._display = False
    oauth = OAuthHelper(reddit, terminal, config)
    assert '.compact' in oauth.reddit.config.API_PATHS['authorize']


def test_oauth_authorize_invalid_token(oauth):

    oauth.config.refresh_token = 'invalid_token'
    oauth.authorize()

    assert isinstance(oauth.term.loader.exception, InvalidRefreshToken)
    assert oauth.server is None
    assert oauth.config.refresh_token is None


def test_oauth_authorize_with_refresh_token(oauth, refresh_token):

    oauth.config.refresh_token = refresh_token
    oauth.authorize(autologin=True)
    assert oauth.server is None

    # We should be able to handle an oauth failure
    with mock.patch.object(oauth.reddit, 'refresh_access_information'):
        exception = OAuthException('', '')
        oauth.reddit.refresh_access_information.side_effect = exception
        oauth.authorize()

    assert isinstance(oauth.term.loader.exception, InvalidRefreshToken)
    assert oauth.server is None
    assert oauth.config.refresh_token is None


def test_oauth_authorize_without_autologin(oauth, terminal, refresh_token):

    # The welcome message should be displayed when autologin is set to
    # false, even if we're using an existing refresh token and not performing
    # the whole login procedure.
    oauth.config.refresh_token = refresh_token
    oauth.authorize(autologin=False)

    text = 'Welcome civilization_phaze_3!'.encode('utf-8')
    terminal.stdscr.subwin.addstr.assert_any_call(1, 1, text)


def test_oauth_clear_data(oauth):
    oauth.config.refresh_token = 'secrettoken'
    oauth.reddit.refresh_token = 'secrettoken'
    oauth.clear_oauth_data()
    assert oauth.config.refresh_token is None
    assert oauth.reddit.refresh_token is None


def test_oauth_authorize(oauth, reddit, stdscr, refresh_token):

    # Because we use `from .helpers import open_browser` we have to patch the
    # function in the destination oauth module and not the helpers module
    with mock.patch('uuid.UUID.hex', new_callable=mock.PropertyMock) as uuid, \
            mock.patch('tuir.terminal.Terminal.open_browser') as open_browser, \
            mock.patch('tuir.oauth.OAuthHTTPServer') as http_server,           \
            mock.patch.object(oauth.reddit, 'user'),                          \
            mock.patch('time.sleep'):

        # Valid authorization
        oauth.term._display = False
        params = {'state': 'uniqueid', 'code': 'secretcode', 'error': None}
        uuid.return_value = params['state']

        def serve_forever():
            oauth.params.update(**params)
        http_server.return_value.serve_forever.side_effect = serve_forever

        oauth.authorize()
        assert open_browser.called
        oauth.reddit.get_access_information.assert_called_with(
            reddit, params['code'])
        assert oauth.config.refresh_token is not None
        assert oauth.config.save_refresh_token.called

        stdscr.reset_mock()
        oauth.reddit.get_access_information.reset_mock()
        oauth.config.save_refresh_token.reset_mock()
        oauth.server = None

        # The next authorization should skip the oauth process
        oauth.config.refresh_token = refresh_token
        oauth.authorize()
        assert oauth.reddit.user is not None
        assert oauth.server is None
        stdscr.reset_mock()

        # Invalid state returned
        params = {'state': 'uniqueid', 'code': 'secretcode', 'error': None}
        oauth.config.refresh_token = None
        uuid.return_value = 'invalidcode'
        oauth.authorize()
        error_message = 'UUID mismatch'.encode('utf-8')
        stdscr.subwin.addstr.assert_any_call(1, 1, error_message)

        # Valid authorization, terminal browser
        oauth.term._display = True
        params = {'state': 'uniqueid', 'code': 'secretcode', 'error': None}
        uuid.return_value = params['state']

        oauth.authorize()
        assert open_browser.called
        oauth.reddit.get_access_information.assert_called_with(
            reddit, params['code'])
        assert oauth.config.refresh_token is not None
        assert oauth.config.save_refresh_token.called

        stdscr.reset_mock()
        oauth.reddit.get_access_information.reset_mock()
        oauth.config.refresh_token = None
        oauth.config.save_refresh_token.reset_mock()
        oauth.server = None

        # Exceptions when logging in are handled correctly
        with mock.patch.object(oauth.reddit, 'get_access_information'):
            exception = OAuthException('', '')
            oauth.reddit.get_access_information.side_effect = exception
            oauth.authorize()
        assert isinstance(oauth.term.loader.exception, OAuthException)
        assert not oauth.config.save_refresh_token.called



from tuir import packages


def test_praw3_package():
    # Sanity check that the package was installed
    assert packages.praw
    assert len(packages.__praw_hash__) == 40
    assert packages.__praw_bundled__ is True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import time
import curses
from collections import OrderedDict

import os
import six
import pytest
import requests
from six.moves import reload_module

from tuir import exceptions
from tuir.objects import Controller, Navigator, Command, KeyMap, \
    curses_session, patch_webbrowser, replace_domain, preferred_reddit_domain

try:
    from unittest import mock
except ImportError:
    import mock

# webbrowser's command to check if a file exists is different for py2/py3
if six.PY3:
    mock_isfile = mock.patch('shutil.which', return_value=None)
else:
    mock_isfile = mock.patch('os.path.isfile', return_value=None)


@mock.patch.dict(os.environ, {'BROWSER': 'safari'})
@mock.patch('sys.platform', 'darwin')
@mock_isfile
def test_patch_webbrowser(*_):

    # Make sure that webbrowser re-generates the browser list using the
    # mocked environment
    import webbrowser
    webbrowser = reload_module(webbrowser)

    # By default, we expect that BROWSER will be loaded as a generic browser
    # This is because "safari" is not a valid script in the system PATH
    assert isinstance(webbrowser.get(), webbrowser.GenericBrowser)

    # After patching, the default webbrowser should now be interpreted as an
    # OSAScript browser
    patch_webbrowser()
    assert isinstance(webbrowser.get(), webbrowser.MacOSXOSAScript)
    assert webbrowser._tryorder[0] == 'safari'


@pytest.mark.parametrize('url,old,new,expected', [
    ('http://old.reddit.com/', 'www.reddit.com', 'old.reddit.com', 'http://www.reddit.com/'),
    ('https://www.reddit.com/', 'old.reddit.com', 'www.reddit.com', 'https://old.reddit.com/'),
    ('https://www.reddit.com/', 'old.reddit.com', 'reddit.com', 'https://www.reddit.com/'),
])
def test_replace_domain(url, new, old, expected):
    assert replace_domain(url, old, new) == expected


@pytest.mark.parametrize('url,old,expected', [
    ('http://old.reddit.com/u/me', False, 'http://old.reddit.com/u/me'),
    ('http://old.reddit.com/u/me', True, 'http://old.reddit.com/u/me'),
    ('https://www.reddit.com/u/me', False, 'https://www.reddit.com/u/me'),
    ('https://www.reddit.com/u/me', True, 'https://old.reddit.com/u/me'),
    ('https://gitlab.com/Chocimier/tuir/', False, 'https://gitlab.com/Chocimier/tuir/'),
    ('https://gitlab.com/Chocimier/tuir/', True, 'https://gitlab.com/Chocimier/tuir/'),
])
def test_preferred_reddit_domain(url, old, expected):
    assert preferred_reddit_domain(url, dict(copy_as_old=old)) == expected


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    # Ensure the thread is properly started/stopped
    with terminal.loader(delay=0, message=u'Hello', trail=u'...'):
        assert terminal.loader._animator.is_alive()
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert terminal.loader.exception is None
    assert stdscr.subwin.ncols == 10
    assert stdscr.subwin.nlines == 3


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_exception_unhandled(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    # Raising an exception should clean up the loader properly
    with pytest.raises(Exception):
        with terminal.loader(delay=0):
            assert terminal.loader._animator.is_alive()
            raise Exception()
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_exception_handled(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    # Raising a handled exception should get stored on the loaders
    with terminal.loader(delay=0):
        assert terminal.loader._animator.is_alive()
        raise requests.ConnectionError()
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert isinstance(terminal.loader.exception, requests.ConnectionError)
    error_message = 'ConnectionError'.encode('ascii' if use_ascii else 'utf-8')
    stdscr.subwin.addstr.assert_called_with(1, 1, error_message)


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_exception_not_caught(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    with pytest.raises(KeyboardInterrupt):
        with terminal.loader(delay=0, catch_exception=False):
            assert terminal.loader._animator.is_alive()
            raise KeyboardInterrupt()
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert terminal.loader.exception is None


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_keyboard_interrupt(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    # Raising a KeyboardInterrupt should be also be stored
    with terminal.loader(delay=0):
        assert terminal.loader._animator.is_alive()
        raise KeyboardInterrupt()
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert isinstance(terminal.loader.exception, KeyboardInterrupt)


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_escape(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    stdscr.getch.return_value = terminal.ESCAPE

    # Pressing escape should trigger an interrupt during the delay section
    with mock.patch('os.kill') as kill:
        with terminal.loader():
            time.sleep(0.1)
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert kill.called

    # As will as during the animation section
    with mock.patch('os.kill') as kill:
        with terminal.loader(delay=0):
            time.sleep(0.1)
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    assert kill.called


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_initial_delay(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    # If we don't reach the initial delay nothing should be drawn
    with terminal.loader(delay=0.1):
        time.sleep(0.05)
    assert not stdscr.subwin.addstr.called


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_nested(terminal, use_ascii):
    terminal.config['ascii'] = use_ascii

    with terminal.loader(message='Outer'):
        with terminal.loader(message='Inner'):
            raise requests.ConnectionError()
        assert False  # Should never be reached

    assert isinstance(terminal.loader.exception, requests.ConnectionError)
    assert terminal.loader.depth == 0
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()


@pytest.mark.parametrize('use_ascii', [True, False])
def test_objects_load_screen_nested_complex(terminal, stdscr, use_ascii):
    terminal.config['ascii'] = use_ascii

    with terminal.loader(message='Outer') as outer_loader:
        assert outer_loader.depth == 1

        with terminal.loader(message='Inner') as inner_loader:
            assert inner_loader.depth == 2
            assert inner_loader._args[2] == 'Outer'

        with terminal.loader():
            assert terminal.loader.depth == 2
            raise requests.ConnectionError()

        assert False  # Should never be reached

    assert isinstance(terminal.loader.exception, requests.ConnectionError)
    assert terminal.loader.depth == 0
    assert not terminal.loader._is_running
    assert not terminal.loader._animator.is_alive()
    error_message = 'ConnectionError'.encode('ascii' if use_ascii else 'utf-8')
    stdscr.subwin.addstr.assert_called_once_with(1, 1, error_message)


def test_objects_curses_session(stdscr):

    # Normal setup and cleanup
    with curses_session():
        pass
    assert curses.initscr.called
    assert curses.endwin.called
    assert curses.use_default_colors.called
    curses.initscr.reset_mock()
    curses.endwin.reset_mock()

    # Ensure cleanup runs if an error occurs
    with pytest.raises(KeyboardInterrupt):
        with curses_session():
            raise KeyboardInterrupt()
    assert curses.initscr.called
    assert curses.endwin.called
    curses.initscr.reset_mock()
    curses.endwin.reset_mock()

    # But cleanup shouldn't run if stdscr was never instantiated
    curses.initscr.side_effect = KeyboardInterrupt
    with pytest.raises(KeyboardInterrupt):
        with curses_session():
            pass
    assert curses.initscr.called
    assert not curses.endwin.called
    curses.initscr.reset_mock()
    curses.endwin.reset_mock()


def test_objects_controller():

    class ControllerA(Controller):
        character_map = {}

    class ControllerB(ControllerA):
        character_map = {}

    class ControllerC(ControllerA):
        character_map = {}

    @ControllerA.register('1')
    def call_page(_):
        return 'a1'

    @ControllerA.register('2')
    def call_page(_):
        return 'a2'

    @ControllerB.register('1')
    def call_page(_):
        return 'b1'

    @ControllerC.register('2')
    def call_page(_):
        return 'c2'

    controller_a = ControllerA(None)
    controller_b = ControllerB(None)
    controller_c = ControllerC(None)

    assert controller_a.trigger('1') == 'a1'
    assert controller_a.trigger('2') == 'a2'
    assert controller_a.trigger('3') is None

    assert controller_b.trigger('1') == 'b1'
    assert controller_b.trigger('2') == 'a2'
    assert controller_b.trigger('3') is None

    assert controller_c.trigger('1') == 'a1'
    assert controller_c.trigger('2') == 'c2'
    assert controller_c.trigger('3') is None


def test_objects_controller_double_press():

    class ControllerA(Controller):
        character_map = {}

    @ControllerA.register(Command('F1'))
    def call_page(_):
        return '1'

    @ControllerA.register(Command('F2'))
    def call_page(_):
        return '2'

    keymap = KeyMap({'F1': ['gg'], 'F2': ['a']})
    controller_a = ControllerA(None, keymap=keymap)

    # Double press
    controller_a.last_char = None
    assert controller_a.trigger('g') is None
    assert controller_a.trigger('g') == '1'
    assert controller_a.trigger('g') is None

    # Executing another command cancels the chain
    controller_a.last_char = None
    assert controller_a.trigger('g') is None
    assert controller_a.trigger('a') == '2'
    assert controller_a.trigger('g') is None

    # Pressing an invalid key cancels the chain
    controller_a.last_char = None
    assert controller_a.trigger('g') is None
    assert controller_a.trigger('b') is None
    assert controller_a.trigger('g') is None


def test_objects_controller_command():

    class ControllerA(Controller):
        character_map = {}

    class ControllerB(ControllerA):
        character_map = {}

    @ControllerA.register(Command('REFRESH'))
    def call_page(_):
        return 'a1'

    @ControllerA.register(Command('UPVOTE'))
    def call_page(_):
        return 'a2'

    @ControllerB.register(Command('REFRESH'))
    def call_page(_):
        return 'b1'

    # Two commands aren't allowed to share keys
    keymap = KeyMap({'REFRESH': [0x10, 0x11], 'UPVOTE': [0x11, 0x12]})
    with pytest.raises(exceptions.ConfigError) as e:
        ControllerA(None, keymap=keymap)
        assert 'ControllerA' in six.text_type(e)

    # Reset the character map
    ControllerA.character_map = {Command('REFRESH'): 0, Command('UPVOTE'): 0}

    # A double command can't share the first key with a single comand
    keymap = KeyMap(OrderedDict([('REFRESH', ['gg']), ('UPVOTE', ['g'])]))
    with pytest.raises(exceptions.ConfigError) as e:
        ControllerA(None, keymap=keymap)
        assert 'ControllerA' in six.text_type(e)

    # It doesn't matter which order they were entered
    keymap = KeyMap(OrderedDict([('UPVOTE', ['g']), ('REFRESH', ['gg'])]))
    with pytest.raises(exceptions.ConfigError) as e:
        ControllerA(None, keymap=keymap)
        assert 'ControllerA' in six.text_type(e)

    # Reset the character map
    ControllerA.character_map = {Command('REFRESH'): 0, Command('UPVOTE'): 0}

    # All commands must be defined in the keymap
    keymap = KeyMap({'REFRESH': [0x10]})
    with pytest.raises(exceptions.ConfigError) as e:
        ControllerB(None, keymap=keymap)
        assert 'UPVOTE' in six.text_type(e)


def test_objects_command():

    c1 = Command("REFRESH")
    c2 = Command("refresh")
    c3 = Command("EXIT")

    assert c1 == c2
    assert c1 != c3

    keymap = {c1: None, c2: None, c3: None}
    assert len(keymap) == 2
    assert c1 in keymap
    assert c2 in keymap
    assert c3 in keymap


def test_objects_keymap():

    bindings = {
        'refresh': ['a', 0x12, '<LF>', '<KEY_UP>'],
        'exit': [],
        Command('UPVOTE'): ['b', '<KEY_F5>'],
        Command('PAGE_TOP'): ['gg']
    }

    keymap = KeyMap(bindings)
    assert keymap.get(Command('REFRESH')) == ['a', 0x12, '<LF>', '<KEY_UP>']
    assert keymap.get(Command('exit')) == []
    assert keymap.get(Command('PAGE_TOP')) == ['gg']
    assert keymap.get('upvote') == ['b', '<KEY_F5>']

    with pytest.raises(exceptions.ConfigError) as e:
        keymap.get('downvote')
        assert 'DOWNVOTE' in six.text_type(e)

    bindings = {'refresh': ['a', 0x12, '<LF>', '<KEY_UP>']}
    bindings2 = {'upvote': ['b', 0x13, '<KEY_DOWN>']}
    keymap._keymap = {}
    keymap.set_bindings(bindings)
    assert keymap.get('refresh')

    with pytest.raises(exceptions.ConfigError) as e:
        keymap.get('upvote')
        assert 'UPVOTE' in six.text_type(e)

    keymap.set_bindings(bindings2)
    assert keymap.get('refresh')
    assert keymap.get('upvote')

    # Strings should be parsed correctly into keys
    assert KeyMap.parse('a') == 97
    assert KeyMap.parse(0x12) == 18
    assert KeyMap.parse('<LF>') == 10
    assert KeyMap.parse('<KEY_UP>') == 259
    assert KeyMap.parse('<KEY_F5>') == 269
    assert KeyMap.parse('gg') == (103, 103)

    for key in ('', None, '<lf>', '<DNS>', '<KEY_UD>', '♬', 'ggg'):
        with pytest.raises(exceptions.ConfigError) as e:
            keymap.parse(key)
            assert six.text_type(key) in six.text_type(e)


def test_objects_navigator_properties():

    def valid_page_cb(_):
        return

    nav = Navigator(valid_page_cb)
    assert nav.step == 1
    assert nav.position == (0, 0, False)
    assert nav.absolute_index == 0
    assert nav.top_item_height is None

    nav = Navigator(valid_page_cb, 5, 2, True, 10)
    assert nav.step == -1
    assert nav.position == (5, 2, True)
    assert nav.absolute_index == 3
    assert nav.top_item_height == 10


def test_objects_navigator_move():

    def valid_page_cb(index):
        if index < 0 or index > 3:
            raise IndexError()

    nav = Navigator(valid_page_cb)
    nav.top_item_height = 5

    # Try to scroll up past the first item
    valid, redraw = nav.move(-1, 2)
    assert not valid
    assert not redraw

    # Scroll down
    valid, redraw = nav.move(1, 3)
    assert nav.page_index == 0
    assert nav.cursor_index == 1
    assert valid
    assert not redraw

    # Scroll down, reach last item on the page and flip the screen
    valid, redraw = nav.move(1, 3)
    assert nav.top_item_height is None
    assert nav.page_index == 2
    assert nav.cursor_index == 0
    assert nav.inverted
    assert valid
    assert redraw

    # Keep scrolling
    valid, redraw = nav.move(1, 3)
    assert nav.page_index == 3
    assert nav.cursor_index == 0
    assert nav.inverted
    assert valid
    assert redraw

    # Reach the end of the page and stop
    valid, redraw = nav.move(1, 1)
    assert nav.page_index == 3
    assert nav.cursor_index == 0
    assert nav.inverted
    assert not valid
    assert not redraw

    # Last item was large and takes up the whole screen, scroll back up and
    # flip the screen again
    valid, redraw = nav.move(-1, 1)
    assert nav.page_index == 2
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert valid
    assert redraw


def test_objects_navigator_move_new_submission():

    def valid_page_cb(index):
        if index != -1:
            raise IndexError()

    nav = Navigator(valid_page_cb, page_index=-1)

    # Can't move up
    valid, redraw = nav.move(-1, 1)
    assert nav.page_index == -1
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert not valid
    assert not redraw

    # Can't move down
    valid, redraw = nav.move(1, 1)
    assert nav.page_index == -1
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert not valid
    assert not redraw


def test_objects_navigator_move_submission():

    def valid_page_cb(index):
        if index < -1 or index > 4:
            raise IndexError()

    nav = Navigator(valid_page_cb, page_index=-1)

    # Can't move up
    valid, redraw = nav.move(-1, 2)
    assert nav.page_index == -1
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert not valid
    assert not redraw

    # Moving down jumps to the first comment
    valid, redraw = nav.move(1, 2)
    assert nav.page_index == 0
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert valid
    assert redraw

    # Moving down again inverts the screen
    valid, redraw = nav.move(1, 2)
    assert nav.page_index == 1
    assert nav.cursor_index == 0
    assert nav.inverted
    assert valid
    assert redraw

    # Move up to the first comment
    valid, redraw = nav.move(-1, 2)
    assert nav.page_index == 0
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert valid
    assert redraw

    # Move up to the submission
    valid, redraw = nav.move(-1, 2)
    assert nav.page_index == -1
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert valid
    assert redraw


@pytest.mark.xfail(reason="Paging is still broken in several edge-cases")
def test_objects_navigator_move_page():

    def valid_page_cb(index):
        if index < 0 or index > 7:
            raise IndexError()

    nav = Navigator(valid_page_cb, cursor_index=2)

    # Can't move up
    valid, redraw = nav.move_page(-1, 5)
    assert nav.page_index == 0
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert not valid
    assert not redraw

    # Page down
    valid, redraw = nav.move_page(1, 5)
    assert nav.page_index == 4
    assert nav.cursor_index == 0
    assert nav.inverted
    assert valid
    assert redraw

    # Page up
    valid, redraw = nav.move_page(-1, 3)
    assert nav.page_index == 2
    assert nav.cursor_index == 0
    assert not nav.inverted
    assert valid
    assert redraw


def test_objects_navigator_flip():

    def valid_page_cb(index):
        if index < 0 or index > 10:
            raise IndexError()

    nav = Navigator(valid_page_cb)

    nav.flip(5)
    assert nav.page_index == 5
    assert nav.cursor_index == 5
    assert nav.inverted

    nav.flip(3)
    assert nav.page_index == 2
    assert nav.cursor_index == 3
    assert not nav.inverted



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import curses

import pytest

from tuir.subscription_page import SubscriptionPage

try:
    from unittest import mock
except ImportError:
    import mock


def test_subscription_page_construct(reddit, terminal, config, oauth,
                                     refresh_token):
    window = terminal.stdscr.subwin

    # Log in
    config.refresh_token = refresh_token
    oauth.authorize()

    with terminal.loader():
        page = SubscriptionPage(reddit, terminal, config, oauth, 'popular')
    assert terminal.loader.exception is None

    page.draw()

    # Header - Title
    title = 'Popular Subreddits'.encode('utf-8')
    window.addstr.assert_any_call(0, 0, title)

    # Header - Name
    name = reddit.user.name.encode('utf-8')
    assert name in [args[0][2] for args in window.addstr.call_args_list]

    # Banner shouldn't be drawn
    menu = '[1]hot     [2]top     [3]rising     [4]new     [5]controversial'.encode('utf-8')
    with pytest.raises(AssertionError):
        window.addstr.assert_any_call(0, 0, menu)

    # Cursor - 2 lines
    window.subwin.addch.assert_any_call(0, 0, ' ', 262144)
    window.subwin.addch.assert_any_call(1, 0, ' ', 262144)

    # Reload with a smaller terminal window
    terminal.stdscr.ncols = 20
    terminal.stdscr.nlines = 10
    with terminal.loader():
        page = SubscriptionPage(reddit, terminal, config, oauth, 'popular')
    assert terminal.loader.exception is None

    page.draw()


def test_subscription_refresh(subscription_page):

    with mock.patch('tuir.terminal.Terminal.flash') as flash:
        # Refresh content - invalid order
        subscription_page.refresh_content(order='top')
        assert flash.called
        flash.reset_mock()

        # Refresh content
        subscription_page.controller.trigger('r')
        assert not flash.called


def test_subscription_prompt(subscription_page, terminal):

    # Prompt for a different subreddit
    with mock.patch.object(terminal, 'prompt_input'):
        # Valid input
        subscription_page.active = True
        subscription_page.selected_page = None
        terminal.prompt_input.return_value = 'front/top'
        subscription_page.controller.trigger('/')

        subscription_page.handle_selected_page()
        assert not subscription_page.active
        assert subscription_page.selected_page

        # Invalid input
        subscription_page.active = True
        subscription_page.selected_page = None
        terminal.prompt_input.return_value = 'front/pot'
        subscription_page.controller.trigger('/')

        subscription_page.handle_selected_page()
        assert subscription_page.active
        assert not subscription_page.selected_page


def test_subscription_move(subscription_page):

    # Test movement
    with mock.patch.object(subscription_page, 'clear_input_queue'):

        # Move cursor down for a little while
        for _ in range(50):
            subscription_page.controller.trigger('j')
        curses.flash.reset_mock()
        assert subscription_page.nav.inverted
        assert (subscription_page.nav.absolute_index ==
                len(subscription_page.content._subscription_data) - 1)

        # And back to the top
        for i in range(subscription_page.nav.absolute_index):
            subscription_page.controller.trigger('k')
        assert not curses.flash.called
        assert subscription_page.nav.absolute_index == 0
        assert not subscription_page.nav.inverted

        # Can't go up any further
        subscription_page.controller.trigger('k')
        assert curses.flash.called
        assert subscription_page.nav.absolute_index == 0
        assert not subscription_page.nav.inverted

        # Page down should move the last item to the top
        n = len(subscription_page._subwindows)
        subscription_page.controller.trigger('n')
        assert subscription_page.nav.absolute_index == n - 1

        # And page up should move back up, but possibly not to the first item
        subscription_page.controller.trigger('m')


def test_subscription_select(subscription_page):

    # Select a subreddit
    subscription_page.controller.trigger(curses.KEY_ENTER)
    subscription_page.handle_selected_page()

    assert subscription_page.selected_page
    assert not subscription_page.active


def test_subscription_close(subscription_page):

    # Close the subscriptions page
    subscription_page.controller.trigger('h')
    assert not subscription_page.selected_page
    assert not subscription_page.active


def test_subscription_page_invalid(subscription_page, oauth, refresh_token):

    oauth.config.refresh_token = refresh_token
    oauth.authorize()

    # Test that other commands don't crash
    methods = [
        'a',  # Upvote
        'z',  # Downvote
        'd',  # Delete
        'e',  # Edit
        'w',  # Save
    ]
    for ch in methods:
        curses.flash.reset_mock()
        subscription_page.controller.trigger(ch)
        assert curses.flash.called



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
from tuir.packages.praw.errors import InvalidUser

from tuir import exceptions
from tuir.docs import FOOTER_INBOX
from tuir.inbox_page import InboxPage
from tuir.submission_page import SubmissionPage

try:
    from unittest import mock
except ImportError:
    import mock


def test_inbox_page_construct(reddit, terminal, config, oauth, refresh_token):
    window = terminal.stdscr.subwin

    # This test assumes the user has at least one message in their inbox
    config.refresh_token = refresh_token
    oauth.authorize()

    with terminal.loader():
        page = InboxPage(reddit, terminal, config, oauth)
    assert terminal.loader.exception is None

    page.draw()

    #  Title
    title = 'My Inbox'.encode('utf-8')
    window.addstr.assert_any_call(0, 0, title)

    # Banner
    menu = '[1]all  [2]unread  [3]messages  [4]comments  [5]posts  [6]mentions  [7]sent'
    window.addstr.assert_any_call(0, 0, menu.encode('utf-8'))

    # Footer - The text is longer than the default terminal width
    text = FOOTER_INBOX.strip()[:79]
    window.addstr.assert_any_call(0, 0, text.encode('utf-8'))

    # Reload with a smaller terminal window
    terminal.stdscr.ncols = 20
    terminal.stdscr.nlines = 10
    with terminal.loader():
        page = InboxPage(reddit, terminal, config, oauth)
    assert terminal.loader.exception is None
    page.draw()


def test_inbox_private_message_parse_file(inbox_page, terminal, reddit):

    with mock.patch.object(terminal, 'open_editor') as open_editor, \
            mock.patch.object(reddit, 'send_message') as send_message, \
            mock.patch('time.sleep'):

        text = ''
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3\n'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3\nsubject'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3\nsubject\n '
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3\n\nbody\n'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        assert not send_message.called

        text = 'civilization_phaze_3\nsubject\nbody\n'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('C')
        send_message.assert_called_with(
            'civilization_phaze_3', 'subject', 'body',
            raise_captcha_exception=True)


def test_inbox_private_message_invalid_author(inbox_page, terminal):

    with mock.patch.object(terminal, 'open_editor') as open_editor:
        text = 'u34891034hjui9oshcvasu7dfashiudhji1293801jdka\nsubject\nbody\n'
        open_editor.return_value.__enter__.return_value = text

        with pytest.raises(exceptions.TemporaryFileError):
            inbox_page.controller.trigger('C')

        assert isinstance(terminal.loader.exception, InvalidUser)


def test_inbox_reply_message(inbox_page, terminal):

    # View messages - this test requires at least one private message
    inbox_page.controller.trigger('3')

    msg = inbox_page.get_selected_item()['object']

    with mock.patch.object(terminal, 'open_editor') as open_editor, \
            mock.patch.object(msg, 'reply') as reply, \
            mock.patch('time.sleep'):

        text = 'My response text'
        open_editor.return_value.__enter__.return_value = text
        inbox_page.controller.trigger('c')
        reply.assert_called_with(text)


def test_inbox_mark_seen(inbox_page, terminal):

    data = inbox_page.get_selected_item()
    message = data['object']

    # Test save on the submission
    with mock.patch.object(message, 'mark_as_read') as mark_as_read, \
            mock.patch.object(message, 'mark_as_unread') as mark_as_unread:

        data['is_new'] = True

        # Mark as seen
        inbox_page.controller.trigger('w')
        assert mark_as_read.called
        assert data['is_new'] is False

        # Mark as no seen
        inbox_page.controller.trigger('w')
        assert mark_as_unread.called
        assert data['is_new'] is True

        # If an exception is raised, state should not be changed
        mark_as_read.side_effect = KeyboardInterrupt
        inbox_page.controller.trigger('w')
        assert data['is_new'] is True


def test_inbox_close(inbox_page, terminal):

    inbox_page.active = None
    inbox_page.controller.trigger('h')
    assert inbox_page.active is False


def test_inbox_view_context(inbox_page, terminal):

    # Should be able to view the context of a comment
    inbox_page.controller.trigger('4')
    inbox_page.controller.trigger('l')
    assert inbox_page.active
    assert inbox_page.selected_page
    assert isinstance(inbox_page.selected_page, SubmissionPage)

    inbox_page.selected_page = None

    # Should not be able to view the context of a private message
    inbox_page.controller.trigger('3')
    inbox_page.controller.trigger('l')
    assert inbox_page.active
    assert inbox_page.selected_page is None
    assert terminal.loader.exception is None


def test_inbox_open_submission(inbox_page, terminal):

    # Should be able to open the submission that a comment was for
    inbox_page.controller.trigger('4')
    inbox_page.controller.trigger('o')
    assert inbox_page.active
    assert inbox_page.selected_page
    assert isinstance(inbox_page.selected_page, SubmissionPage)

    inbox_page.selected_page = None

    # Should not be able to open the submission for a private message
    inbox_page.controller.trigger('3')
    inbox_page.controller.trigger('o')
    assert inbox_page.active
    assert inbox_page.selected_page is None
    assert terminal.loader.exception is None
